from typing import Literal, Optional

from jaxtyping import Array, Bool

from ._config import AbstractConfig, DoseConfig
from ._detector import AbstractDetector
from ._image_model import (
    AbstractImageModel,
    ContrastImageModel,
    ElectronCountsImageModel,
    IntensityImageModel,
    LinearImageModel,
    ProjectionImageModel as ProjectionImageModel,
)
from ._pose import AbstractPose
from ._potential_integrator import (
    AbstractDirectIntegrator,
    FourierSliceExtraction,
    GaussianMixtureProjection,
    NufftProjection,
)
from ._potential_representation import (
    AbstractPotentialRepresentation,
    FourierVoxelGridPotential,
    FourierVoxelSplinePotential,
    GaussianMixtureAtomicPotential,
    PengAtomicPotential,
    RealVoxelCloudPotential,
    RealVoxelGridPotential,
)
from ._scattering_theory import WeakPhaseScatteringTheory
from ._structure import BasicStructure
from ._transfer_theory import ContrastTransferTheory


def make_image_model(
    potential: AbstractPotentialRepresentation,
    config: AbstractConfig,
    pose: AbstractPose,
    transfer_theory: Optional[ContrastTransferTheory] = None,
    integrator: Optional[AbstractDirectIntegrator] = None,
    detector: Optional[AbstractDetector] = None,
    *,
    normalizes_signal: bool = False,
    signal_region: Optional[Bool[Array, "{config.y_dim} {config.x_dim}"]] = None,
    physical_units: bool = True,
    mode: Literal["contrast", "intensity", "counts"] = "contrast",
) -> AbstractImageModel:
    """Construct an `AbstractImageModel` for most common use-cases.

    **Arguments:**

    - `potential`:
        The representation of the protein electrostatic potential.
        Common choices are the `FourierVoxelGridPotential`
        for fourier-space voxel grids or the `PengAtomicPotential`
        for gaussian mixtures of atoms parameterized by electron scattering factors.
    - `config`:
        The configuration for the image and imagining instrument. Unless using
        a model that uses the electron dose as a parameter, choose the
        `InstrumentConfig`. Otherwise, choose the `DoseConfig`.
    - `pose`:
        The pose in a particular parameterization convention. Common options
        are the `EulerAnglePose`, `QuaternionPose`, or `AxisAnglePose`.
    - `transfer_theory`:
        The contrast transfer function and its theory for how it is applied
        to the image.
    - `integrator`:
        Optionally pass the method for integrating the electrostatic potential onto
        the plane (e.g. projection via fourier slice extraction). If not provided,
        a default option is chosen.
    - `detector`:
        If `mode = 'counts'` is chosen, then an `AbstractDetector` class must be
        chosen to simulate electron counts.
    - `normalizes_signal`:
        Whether or not to normalize the image.
    - `signal_region`:
        If `normalizes_signal = True`, this is a boolean array that is 1 where
        there is signal and 0 otherwise.
    - `physical_units`:
        If `True`, the image simulated is a physical quantity, which is
        chosen with the `mode` argument. Otherwise, simulate an image without
        scaling to absolute units.
    - `mode`:
        The physical observable to simulate. Not used if `physical_units = False`.
        Options are
        - 'contrast':
            Uses the `ContrastImageModel` to simulate contrast. This is
            default.
        - 'intensity':
            Uses the `IntensityImageModel` to simulate intensity.
        - 'counts':
            Uses the `ElectronCountsImageModel` to simulate electron counts.
            If this is passed, a `detector` must also be passed.

    **Returns:**

    An `AbstractImageModel`. Simulate an image with syntax

    ```python
    image_model = make_image_model(...)
    image = image_model.simulate()
    ```
    """
    # Build the image model
    integrator = _select_default_integrator(potential)
    structure = BasicStructure(potential, pose)
    if transfer_theory is None:
        image_model = ProjectionImageModel(
            structure,
            config,
            integrator,
            normalizes_signal=normalizes_signal,
            signal_region=signal_region,
        )
    else:
        if physical_units:
            scattering_theory = WeakPhaseScatteringTheory(integrator, transfer_theory)
            if mode == "counts":
                if not isinstance(config, DoseConfig):
                    raise ValueError(
                        "If using `mode = 'counts'` to simulate electron counts, "
                        "pass `config = DoseConfig(...)`. Got config "
                        f"{type(config).__name__}."
                    )
                if detector is None:
                    raise ValueError(
                        "If using `mode = 'counts'` to simulate electron counts, "
                        "an `AbstractDetector` must be passed."
                    )
                image_model = ElectronCountsImageModel(
                    structure,
                    config,
                    scattering_theory,
                    detector,
                    normalizes_signal=normalizes_signal,
                    signal_region=signal_region,
                )
            elif mode == "contrast":
                image_model = ContrastImageModel(
                    structure,
                    config,
                    scattering_theory,
                    normalizes_signal=normalizes_signal,
                    signal_region=signal_region,
                )
            elif mode == "intensity":
                image_model = IntensityImageModel(
                    structure,
                    config,
                    scattering_theory,
                    normalizes_signal=normalizes_signal,
                    signal_region=signal_region,
                )
            else:
                raise ValueError(
                    f"`mode = {mode}` not supported. Supported modes for simulating "
                    "physical quantities are 'contrast', 'intensity', and 'counts'."
                )
        else:
            image_model = LinearImageModel(
                structure,
                config,
                integrator,
                transfer_theory,
                normalizes_signal=normalizes_signal,
                signal_region=signal_region,
            )

    return image_model


def _select_default_integrator(
    potential: AbstractPotentialRepresentation,
) -> AbstractDirectIntegrator:
    if isinstance(potential, (FourierVoxelGridPotential, FourierVoxelSplinePotential)):
        integrator = FourierSliceExtraction()
    elif isinstance(potential, (PengAtomicPotential, GaussianMixtureAtomicPotential)):
        integrator = GaussianMixtureProjection(use_error_functions=True)
    elif isinstance(potential, (RealVoxelCloudPotential, RealVoxelGridPotential)):
        integrator = NufftProjection()
    else:
        raise ValueError(
            "Could not select default integrator for potential of "
            f"type {type(potential).__name__}. If using a custom potential "
            "please directly pass an integrator."
        )
    return integrator



"""
Base scattering potential representation.
"""

from abc import abstractmethod
from typing_extensions import Self

from equinox import Module

from .._pose import AbstractPose


class AbstractPotentialRepresentation(Module, strict=True):
    """Abstract interface for the spatial potential energy distribution of a
    scatterer.
    """

    @abstractmethod
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new `AbstractPotentialRepresentation` at the given pose.

        **Arguments:**

        - `pose`: The pose at which to view the `AbstractPotentialRepresentation`.
        """
        raise NotImplementedError



from ._dataset import AbstractDataset as AbstractDataset
from ._particle_data import (
    AbstractParticleParameterFile as AbstractParticleParameterFile,
    AbstractParticleStackDataset as AbstractParticleStackDataset,
    AbstractParticleStarFile as AbstractParticleStarFile,
    RelionParticleParameterFile as RelionParticleParameterFile,
    RelionParticleStackDataset as RelionParticleStackDataset,
    simulate_particle_stack as simulate_particle_stack,
)



"""Implementation of an AbstractFourierOperator. Put simply, these are
functions commonly applied to images in fourier space.

Opposed to a AbstractFilter, a AbstractFourierOperator is computed at
runtime---not upon initialization. AbstractFourierOperators also do not
have a rule for how they should be applied to images.

These classes are modified from the library ``tinygp``.
"""

from abc import abstractmethod
from typing import overload
from typing_extensions import override

import jax.numpy as jnp
from jaxtyping import Array, Float, Inexact

from ...internal import error_if_negative, error_if_not_positive
from ._base_operator import AbstractImageOperator


class AbstractFourierOperator(AbstractImageOperator, strict=True):
    """
    The base class for all fourier-based operators.

    By convention, operators should be defined to
    be dimensionless (up to a scale factor).

    To create a subclass,

        1) Include the necessary parameters in
           the class definition.
        2) Overrwrite the ``__call__`` method.
    """

    @overload
    @abstractmethod
    def __call__(
        self, frequency_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Inexact[Array, "y_dim x_dim"]: ...

    @overload
    @abstractmethod
    def __call__(  # type: ignore
        self, frequency_grid: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Inexact[Array, "z_dim y_dim x_dim"]: ...

    @abstractmethod
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        raise NotImplementedError


FourierOperatorLike = AbstractFourierOperator | AbstractImageOperator


class ZeroMode(AbstractFourierOperator, strict=True):
    """This operator returns a constant in the zero mode."""

    value: Float[Array, ""]

    def __init__(self, value: float | Float[Array, ""] = 0.0):
        """**Arguments:**

        - `value`: The value of the zero mode.
        """
        self.value = jnp.asarray(value)

    @override
    def __call__(
        self, frequency_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        N1, N2 = frequency_grid.shape[0:-1]
        return jnp.zeros((N1, N2)).at[0, 0].set(self.value)


class FourierExp2D(AbstractFourierOperator, strict=True):
    r"""This operator, in real space, represents a
    function equal to an exponential decay, given by

    .. math::
        g(|r|) = \frac{\kappa}{2 \pi \xi^2} \exp(- |r| / \xi),

    where :math:`|r| = \sqrt{x^2 + y^2}` is a radial coordinate.
    Here, :math:`\xi` has dimensions of length and :math:`g(r)`
    has dimensions of inverse area. The power spectrum from such
    a correlation function (in two-dimensions) is given by its
    Hankel transform pair

    .. math::
        P(|k|) = \frac{\kappa}{2 \pi \xi^3} \frac{1}{(\xi^{-2} + |k|^2)^{3/2}}.

    Here :math:`\kappa` is a scale factor and :math:`\xi` is a length
    scale.
    """

    amplitude: Float[Array, ""]
    length_scale: Float[Array, ""]

    def __init__(
        self,
        amplitude: float | Float[Array, ""] = 1.0,
        length_scale: float | Float[Array, ""] = 1.0,
    ):
        """**Arguments:**

        - `amplitude`: The amplitude of the operator, equal to $\\kappa$
                in the above equation.
        - `length_scale`: The length scale of the operator, equal to $\\xi$
                    in the above equation.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.length_scale = error_if_not_positive(jnp.asarray(length_scale, dtype=float))

    @override
    def __call__(
        self, frequency_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        k_sqr = jnp.sum(frequency_grid**2, axis=-1)
        scaling = 1.0 / (k_sqr + jnp.divide(1, (self.length_scale) ** 2)) ** 1.5
        scaling *= jnp.divide(self.amplitude, 2 * jnp.pi * self.length_scale**3)
        return scaling


class Lorenzian(AbstractFourierOperator, strict=True):
    r"""This operator is the Lorenzian, given

    .. math::
        P(|k|) = \frac{\kappa}{\xi^2} \frac{1}{(\xi^{-2} + |k|^2)}.

    Here :math:`\kappa` is a scale factor and :math:`\xi` is a length
    scale.
    """

    amplitude: Float[Array, ""]
    length_scale: Float[Array, ""]

    def __init__(
        self,
        amplitude: float | Float[Array, ""] = 1.0,
        length_scale: float | Float[Array, ""] = 1.0,
    ):
        """**Arguments:**

        - `amplitude`: The amplitude of the operator, equal to $\\kappa$
                in the above equation.
        - `length_scale`: The length scale of the operator, equal to $\\xi$
                    in the above equation.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.length_scale = error_if_not_positive(jnp.asarray(length_scale, dtype=float))

    @overload
    def __call__(
        self, frequency_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]: ...

    @overload
    def __call__(  # type: ignore
        self, frequency_grid: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "z_dim y_dim x_dim"]: ...

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        k_sqr = jnp.sum(frequency_grid**2, axis=-1)
        scaling = 1.0 / (k_sqr + jnp.divide(1, self.length_scale**2))
        scaling *= jnp.divide(self.amplitude, self.length_scale**2)
        return scaling


class FourierGaussian(AbstractFourierOperator, strict=True):
    r"""This operator represents a simple gaussian.
    Specifically, this is

    .. math::
        P(k) = \kappa \exp(- \beta k^2 / 4),

    where :math:`k^2 = k_x^2 + k_y^2` is the length of the
    wave vector. Here, :math:`\beta` has dimensions of length
    squared.
    """

    amplitude: Float[Array, ""]
    b_factor: Float[Array, ""]

    def __init__(
        self,
        amplitude: float | Float[Array, ""] = 1.0,
        b_factor: float | Float[Array, ""] = 1.0,
    ):
        """**Arguments:**

        - `amplitude`:
            The amplitude of the operator, equal to $\\kappa$
            in the above equation.
        - `b_factor`:
            The B-factor of the gaussian, equal to $\\beta$
            in the above equation.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.b_factor = error_if_not_positive(jnp.asarray(b_factor, dtype=float))

    @overload
    def __call__(
        self, frequency_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]: ...

    @overload
    def __call__(  # type: ignore
        self, frequency_grid: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "z_dim y_dim x_dim"]: ...

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        k_sqr = jnp.sum(frequency_grid**2, axis=-1)
        scaling = self.amplitude * jnp.exp(-0.25 * self.b_factor * k_sqr)
        return scaling


class FourierGaussianWithRadialOffset(AbstractFourierOperator, strict=True):
    r"""This operator represents a gaussian with a radial offset.
    Specifically, this is

    .. math::
        P(k) = \kappa \exp(- \beta (|k| - m)^2 / 4),

    where :math:`k^2 = k_x^2 + k_y^2` is the length of the
    wave vector. Here, :math:`\beta` has dimensions of length
    squared.
    """

    amplitude: Float[Array, ""]
    b_factor: Float[Array, ""]
    offset: Float[Array, ""]

    def __init__(
        self,
        amplitude: float | Float[Array, ""] = 1.0,
        b_factor: float | Float[Array, ""] = 1.0,
        offset: float | Float[Array, ""] = 0.0,
    ):
        """**Arguments:**

        - `amplitude`:
            The amplitude of the operator, equal to $\\kappa$
            in the above equation.
        - `b_factor`:
            The B-factor of the gaussian, equal to $\\beta$
            in the above equation.
        - `offset`:
            The radial offset of the gaussian.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.b_factor = error_if_not_positive(jnp.asarray(b_factor, dtype=float))
        self.offset = error_if_negative(jnp.asarray(offset, dtype=float))

    @overload
    def __call__(
        self, frequency_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]: ...

    @overload
    def __call__(  # type: ignore
        self, frequency_grid: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "z_dim y_dim x_dim"]: ...

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        k = jnp.linalg.norm(frequency_grid, axis=-1)
        scaling = self.amplitude * jnp.exp(-0.25 * self.b_factor * (k - self.offset) ** 2)
        return scaling



import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, install_import_hook


with install_import_hook("cryojax", "typeguard.typechecked"):
    from cryojax.jax_util import (
        CustomTransform,
        StopGradientTransform,
        resolve_transforms,
    )


class Exp(eqx.Module):
    a: Array = eqx.field(converter=jnp.asarray)

    def __call__(self, x):
        return jnp.exp(-self.a * x)


def test_resolve_transform():
    pytree = Exp(a=1.0)
    pytree_with_transform = eqx.tree_at(
        lambda fn: fn.a,
        pytree,
        replace_fn=lambda a: CustomTransform(jnp.exp, jnp.log(a)),
    )
    assert eqx.tree_equal(pytree, resolve_transforms(pytree_with_transform))


def test_nested_resolve_transform():
    pytree = Exp(a=1.0)
    pytree_with_transform = eqx.tree_at(
        lambda fn: fn.a,
        pytree,
        replace_fn=lambda a: CustomTransform(lambda b: 2 * b, a / 2),
    )
    pytree_with_nested_transform = eqx.tree_at(
        lambda fn: fn.a.args[0],
        pytree_with_transform,
        replace_fn=lambda a_scaled: CustomTransform(jnp.exp, jnp.log(a_scaled)),
    )
    assert eqx.tree_equal(
        pytree,
        resolve_transforms(pytree_with_transform),
        resolve_transforms(pytree_with_nested_transform),
    )


def test_stop_gradient():
    @jax.value_and_grad
    def objective_fn(pytree):
        exp, x = resolve_transforms(pytree)
        return exp(x)

    x = jnp.asarray(np.random.random())
    exp = Exp(a=1.0)
    exp_with_stop_gradient = eqx.tree_at(
        lambda fn: fn.a, exp, replace_fn=StopGradientTransform
    )
    _, grads = objective_fn((exp_with_stop_gradient, x))
    grads = resolve_transforms(grads)
    assert grads[0].a == 0.0



import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact


def apply_interaction_constant(
    integrated_potential: Inexact[Array, "y_dim x_dim"],
    wavelength_in_angstroms: Float[Array, ""] | float,
) -> Inexact[Array, "y_dim x_dim"]:
    """Given an integrated potential, convert units to the object
    phase shift distribution using the interaction
    constant. For example, the case of the projection approximation,
    compute the phase shift distribution.

    !!! info
        In the projection approximation in cryo-EM, the phase shifts in the
        exit plane are given by

        $$\\eta(x, y) = \\sigma \\int dz \\ V(x, y, z),$$

        where $\\sigma$ is typically referred to as the interaction
        constant. However, in `cryojax`, the potential is rescaled
        to units of inverse length squared as

        $$U(x, y, z) = \\frac{2 m e}{\\hbar^2} V(x, y, z).$$

        With this rescaling of the potential, the phase shifts are equal to

        $$\\eta(x, y) = \\frac{\\lambda}{4 \\pi} \\int dz \\ U(x, y, z).$$

        **References**:

        - For the definition of the rescaled potential, see
        Chapter 69, Page 2003, Equation 69.6 from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*
        - For the definition of the phase shifts in terms of the rescaled potential, see
        Chapter 69, Page 2012, Equation 69.34b from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*

    See the documentation on atom-based scattering potentials for more information.
    """  # noqa: E501
    return integrated_potential * jnp.asarray(wavelength_in_angstroms) / (4 * jnp.pi)


def apply_amplitude_contrast_ratio(
    integrated_potential: Float[Array, "y_dim x_dim"] | Complex[Array, "y_dim x_dim"],
    amplitude_contrast_ratio: Float[Array, ""] | float,
) -> Complex[Array, "y_dim x_dim"]:
    """Apply the amplitude contrast ratio to compute a complex scattering potential,
    given the integrated potential computed just with a weak-phase calculation.

    !!! info
        Given a projected electrostatic potential $\\u(x, y)$, the complex potential
        $\\phi(x, y)$ for amplitude contrast ratio $\\alpha$ is

        $$\\phi(x, y) = \\sqrt{1 - \\alpha^2} \\ u(x, y) + i \\alpha \\ u(x, y)$$
    """
    ac = amplitude_contrast_ratio
    if jnp.iscomplexobj(integrated_potential):
        return jnp.sqrt(1.0 - ac**2) * integrated_potential.real + 1.0j * (
            integrated_potential.imag + ac * integrated_potential.real
        )
    else:
        return (jnp.sqrt(1.0 - ac**2) + 1.0j * ac) * integrated_potential



import warnings

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jaxtyping import Array

import cryojax.simulator as cxs
from cryojax.constants import (
    get_tabulated_scattering_factor_parameters,
    read_peng_element_scattering_factor_parameter_table,
)
from cryojax.io import read_atoms_from_pdb
from cryojax.ndimage import crop_to_shape, irfftn


jax.config.update("jax_enable_x64", True)


@pytest.mark.parametrize(
    "pixel_size, shape",
    (
        (1.0, (32, 32)),
        (1.0, (31, 31)),
        (1.0, (31, 32)),
        (1.0, (32, 31)),
    ),
)
def test_projection_methods_no_pose(sample_pdb_path, pixel_size, shape):
    """Test that computing a projection in real
    space agrees with real-space, with no rotation. This mostly
    makes sure there are no numerical artifacts in fourier space
    interpolation and that volumes are read in real vs. fourier
    at the same orientation.
    """
    # Objects for imaging
    config = cxs.BasicConfig(
        shape,
        pixel_size,
        voltage_in_kilovolts=300.0,
    )
    # Real vs fourier potentials
    dim = max(*shape)  # Make sure to use `padded_shape` here
    atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
        sample_pdb_path, center=True, loads_b_factors=True
    )
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_identities, read_peng_element_scattering_factor_parameter_table()
    )
    base_potential = cxs.PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
        b_factors=b_factors,
    )
    base_method = cxs.GaussianMixtureProjection(use_error_functions=True)

    real_voxel_grid = base_potential.as_real_voxel_grid((dim, dim, dim), pixel_size)
    other_potentials = [
        cxs.FourierVoxelGridPotential.from_real_voxel_grid(real_voxel_grid, pixel_size),
        make_spline_potential(real_voxel_grid, pixel_size),
        cxs.GaussianMixtureAtomicPotential(
            atom_positions,
            scattering_factor_parameters["a"],
            (scattering_factor_parameters["b"] + b_factors[:, None]) / (8 * jnp.pi**2),
        ),
        cxs.RealVoxelGridPotential.from_real_voxel_grid(real_voxel_grid, pixel_size),
        cxs.RealVoxelCloudPotential.from_real_voxel_grid(
            real_voxel_grid, pixel_size, rtol=0.0, atol=1e-16
        ),
    ]
    other_projection_methods = [
        cxs.FourierSliceExtraction(),
        cxs.FourierSliceExtraction(),
        base_method,
        cxs.NufftProjection(eps=1e-16),
        cxs.NufftProjection(eps=1e-16),
    ]

    projection_by_gaussian_integration = compute_projection(
        base_potential, base_method, config
    )
    for potential, projection_method in zip(other_potentials, other_projection_methods):
        if isinstance(projection_method, cxs.NufftProjection):
            try:
                projection_by_other_method = compute_projection(
                    potential, projection_method, config
                )
            except Exception as err:
                warnings.warn(
                    "Could not test projection method `NufftProjection` "
                    "This is most likely because `jax_finufft` is not installed. "
                    f"Error traceback is:\n{err}"
                )
                continue
        else:
            projection_by_other_method = compute_projection(
                potential, projection_method, config
            )
        np.testing.assert_allclose(
            projection_by_gaussian_integration, projection_by_other_method, atol=1e-12
        )


# @pytest.mark.parametrize(
#     "pixel_size, shape, euler_pose_params",
#     (
#         (1.0, (32, 32), (2.5, -5.0, 0.0, 0.0, 0.0)),
#         (1.0, (32, 32), (0.0, 0.0, 10.0, -30.0, 60.0)),
#         (1.0, (32, 32), (2.5, -5.0, 10.0, -30.0, 60.0)),
#     ),
# )
# def test_projection_methods_with_pose(
#     sample_pdb_path, pixel_size, shape, euler_pose_params
# ):
#     """Test that computing a projection across different
#     methods agrees. This tests pose convention and accuracy
#     for real vs fourier, atoms vs voxels, etc.
#     """
#     # Objects for imaging
#     instrument_config = cxs.BasicConfig(
#         shape,
#         pixel_size,
#         voltage_in_kilovolts=300.0,
#     )
#     euler_pose = cxs.EulerAnglePose(*euler_pose_params)
#     # Real vs fourier potentials
#     dim = max(*shape)
#     atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
#         sample_pdb_path, center=True, loads_b_factors=True
#     )
#     scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
#         atom_identities, read_peng_element_scattering_factor_parameter_table()
#     )
#     base_potential = cxs.PengAtomicPotential(
#         atom_positions,
#         scattering_factor_a=scattering_factor_parameters["a"],
#         scattering_factor_b=scattering_factor_parameters["b"],
#         b_factors=b_factors,
#     )
#     base_method = cxs.GaussianMixtureProjection(use_error_functions=True)

#     real_voxel_grid = base_potential.as_real_voxel_grid((dim, dim, dim), pixel_size)
#     other_potentials = [
#         cxs.FourierVoxelGridPotential.from_real_voxel_grid(real_voxel_grid, pixel_size),
#         make_spline_potential(real_voxel_grid, pixel_size),
#         cxs.GaussianMixtureAtomicPotential(
#             atom_positions,
#             scattering_factor_parameters["a"],
#             (scattering_factor_parameters["b"] + b_factors[:, None]) / (8 * jnp.pi**2),
#         ),
#     ]
#     #     cxs.RealVoxelGridPotential.from_real_voxel_grid(real_voxel_grid, pixel_size),
#     #     cxs.RealVoxelCloudPotential.from_real_voxel_grid(real_voxel_grid, pixel_size),
#     # ]
#     other_projection_methods = [
#         cxs.FourierSliceExtraction(),
#         cxs.FourierSliceExtraction(),
#         base_method,
#     ]
#     #     cxs.NufftProjection(),
#     #     cxs.NufftProjection(),
#     # ]

#     projection_by_gaussian_integration = compute_projection_at_pose(
#         base_potential, base_method, euler_pose, instrument_config
#     )
#     for idx, (potential, projection_method) in enumerate(
#         zip(other_potentials, other_projection_methods)
#     ):
#         if isinstance(projection_method, cxs.NufftProjection):
#             try:
#                 projection_by_other_method = compute_projection_at_pose(
#                     potential, projection_method, euler_pose, instrument_config
#                 )
#             except Exception as err:
#                 warnings.warn(
#                     "Could not test projection method `NufftProjection` "
#                     "This is most likely because `jax_finufft` is not installed. "
#                     f"Error traceback is:\n{err}"
#                 )
#                 continue
#         else:
#             projection_by_other_method = compute_projection_at_pose(
#                 potential, projection_method, euler_pose, instrument_config
#             )
#         np.testing.assert_allclose(
#             np.sum(
#                 (projection_by_gaussian_integration - projection_by_other_method) ** 2
#             ),
#             0.0,
#             atol=1e-8,
#         )


@eqx.filter_jit
def compute_projection(
    potential: cxs.AbstractPotentialRepresentation,
    integrator: cxs.AbstractPotentialIntegrator,
    config: cxs.BasicConfig,
) -> Array:
    fourier_projection = integrator.integrate(potential, config, outputs_real_space=False)
    return crop_to_shape(
        irfftn(
            fourier_projection,
            s=config.padded_shape,
        ),
        config.shape,
    )


@eqx.filter_jit
def compute_projection_at_pose(
    potential: cxs.AbstractPotentialRepresentation,
    integrator: cxs.AbstractPotentialIntegrator,
    pose: cxs.AbstractPose,
    config: cxs.BasicConfig,
) -> Array:
    rotated_potential = potential.rotate_to_pose(pose)
    fourier_projection = integrator.integrate(
        rotated_potential, config, outputs_real_space=False
    )
    translation_operator = pose.compute_translation_operator(
        config.padded_frequency_grid_in_angstroms
    )
    return crop_to_shape(
        irfftn(
            pose.translate_image(
                fourier_projection,
                translation_operator,
                config.padded_shape,
            ),
            s=config.padded_shape,
        ),
        config.shape,
    )


@eqx.filter_jit
def make_spline_potential(real_voxel_grid, voxel_size):
    return cxs.FourierVoxelSplinePotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size
    )



"""
Implementation of operators on images in real-space.
"""

from abc import abstractmethod
from typing import overload
from typing_extensions import override

import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float

from ...internal import error_if_not_positive
from ._base_operator import AbstractImageOperator


class AbstractRealOperator(AbstractImageOperator, strict=True):
    """
    The base class for all real operators.

    By convention, operators should be defined to
    have units of inverse area (up to a scale factor).

    To create a subclass,

        1) Include the necessary parameters in
           the class definition.
        2) Overrwrite the ``__call__`` method.
    """

    @overload
    @abstractmethod
    def __call__(
        self, coordinate_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]: ...

    @overload
    @abstractmethod
    def __call__(  # type: ignore
        self, coordinate_grid: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "z_dim y_dim x_dim"]: ...

    @abstractmethod
    def __call__(  # pyright: ignore
        self,
        coordinate_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        raise NotImplementedError


RealOperatorLike = AbstractRealOperator | AbstractImageOperator


class Gaussian2D(AbstractRealOperator, strict=True):
    """This operator represents a simple gaussian in 2D.
    Specifically, this is

    $$g(r) = \\frac{\\kappa}{2\\pi \\beta} \\exp(- (r - r_0)^2 / (2 \\sigma))$$

    where $r^2 = x^2 + y^2$.
    """

    amplitude: Float[Array, ""]
    variance: Float[Array, ""]
    offset: Float[Array, "2"]

    def __init__(
        self,
        amplitude: float | Float[Array, ""] = 1.0,
        variance: float | Float[Array, ""] = 1.0,
        offset: Float[Array | np.ndarray, "2"] | tuple[float, float] = (0.0, 0.0),
    ):
        """**Arguments:**

        - `amplitude`:
            The amplitude of the operator, equal to $\\kappa$
            in the above equation.
        - `variance`:
            The variance of the gaussian, equal to $\\sigma$
            in the above equation.
        - `offset`:
            An offset to the origin, equal to $r_0$
            in the above equation.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.variance = error_if_not_positive(jnp.asarray(variance, dtype=float))
        self.offset = jnp.asarray(offset, dtype=float)

    @override
    def __call__(
        self, coordinate_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        r_sqr = jnp.sum((coordinate_grid - self.offset) ** 2, axis=-1)
        scaling = (self.amplitude / jnp.sqrt(2 * jnp.pi * self.variance)) * jnp.exp(
            -0.5 * r_sqr / self.variance
        )
        return scaling



from typing import Optional
from typing_extensions import override

from jaxtyping import Array, Complex, Float, PRNGKeyArray

from .._common_functions import apply_interaction_constant
from .._config import AbstractConfig
from .._potential_integrator import AbstractDirectIntegrator
from .._potential_representation import AbstractPotentialRepresentation
from .._solvent import AbstractRandomSolvent
from .._transfer_theory import ContrastTransferTheory
from .base_scattering_theory import AbstractWeakPhaseScatteringTheory


class WeakPhaseScatteringTheory(AbstractWeakPhaseScatteringTheory, strict=True):
    """Base linear image formation theory."""

    integrator: AbstractDirectIntegrator
    transfer_theory: ContrastTransferTheory
    solvent: Optional[AbstractRandomSolvent] = None

    def __init__(
        self,
        integrator: AbstractDirectIntegrator,
        transfer_theory: ContrastTransferTheory,
        solvent: Optional[AbstractRandomSolvent] = None,
    ):
        """**Arguments:**

        - `integrator`: The method for integrating the scattering potential.
        - `transfer_theory`: The contrast transfer theory.
        - `solvent`: The model for the solvent.
        """
        self.integrator = integrator
        self.transfer_theory = transfer_theory
        self.solvent = solvent

    @override
    def compute_object_spectrum(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        # Compute the integrated potential
        fourier_in_plane_potential = self.integrator.integrate(
            potential, config, outputs_real_space=False
        )

        if rng_key is not None:
            # Get the potential of the specimen plus the ice
            if self.solvent is not None:
                fourier_in_plane_potential = self.solvent.compute_in_plane_potential(  # noqa: E501
                    rng_key,
                    fourier_in_plane_potential,
                    config,
                    input_is_rfft=self.integrator.is_projection_approximation,
                )

        object_spectrum = apply_interaction_constant(
            fourier_in_plane_potential, config.wavelength_in_angstroms
        )

        return object_spectrum

    @override
    def compute_contrast_spectrum(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
        defocus_offset: Optional[float | Float[Array, ""]] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        object_spectrum = self.compute_object_spectrum(potential, config, rng_key)
        contrast_spectrum = self.transfer_theory.propagate_object(  # noqa: E501
            object_spectrum,
            config,
            is_projection_approximation=self.integrator.is_projection_approximation,
            defocus_offset=defocus_offset,
        )

        return contrast_spectrum



import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jax import config
from jaxtyping import Array, Float, install_import_hook

from cryojax.ndimage import downsample_with_fourier_cropping


with install_import_hook("cryojax", "typeguard.typechecked"):
    import cryojax.simulator as cxs
    from cryojax.constants import (
        get_tabulated_scattering_factor_parameters,
        read_peng_element_scattering_factor_parameter_table,
    )
    from cryojax.coordinates import make_coordinate_grid
    from cryojax.io import read_atoms_from_pdb
    from cryojax.ndimage import ifftn
    from cryojax.simulator import (
        FourierVoxelGridPotential,
        GaussianMixtureAtomicPotential,
        PengAtomicPotential,
        RealVoxelGridPotential,
    )


config.update("jax_enable_x64", True)


@pytest.fixture
def toy_gaussian_cloud():
    atom_positions = jnp.array(
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ]
    )
    num_atoms = atom_positions.shape[0]
    ff_a = jnp.array(
        num_atoms
        * [
            [1.0, 0.5],
        ]
    )

    ff_b = jnp.array(
        num_atoms
        * [
            [0.3, 0.2],
        ]
    )

    n_voxels_per_side = (128, 128, 128)
    voxel_size = 0.05
    return (atom_positions, ff_a, ff_b, n_voxels_per_side, voxel_size)


#
# Test different representations
#
def test_voxel_potential_loaders():
    real_voxel_grid = jnp.zeros((10, 10, 10), dtype=float)
    voxel_size = 1.1
    fourier_potential = cxs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    real_potential = cxs.RealVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    cloud_potential = cxs.RealVoxelCloudPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    for potential in [real_potential, fourier_potential, cloud_potential]:
        assert potential.voxel_size == jnp.asarray(voxel_size)

    assert isinstance(
        fourier_potential.frequency_slice_in_pixels,
        Float[Array, "1 _ _ 3"],  # type: ignore
    )
    assert isinstance(real_potential.coordinate_grid_in_pixels, Float[Array, "_ _ _ 3"])  # type: ignore
    assert isinstance(cloud_potential.coordinate_list_in_pixels, Float[Array, "_ 3"])  # type: ignore


#
# Test rendering
#
def test_fourier_vs_real_voxel_potential_agreement(sample_pdb_path):
    """
    Integration test ensuring that the VoxelGrid classes
    produce comparable electron densities when loaded from PDB.
    """
    n_voxels_per_side = (128, 128, 128)
    voxel_size = 0.5

    # Load the PDB file
    atom_positions, atom_elements = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        loads_b_factors=False,
        selection_string="not element H",
    )
    # Load atomistic potential
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_elements, read_peng_element_scattering_factor_parameter_table()
    )
    atomic_potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
    )
    # Build the grid
    potential_as_real_voxel_grid = atomic_potential.as_real_voxel_grid(
        n_voxels_per_side, voxel_size
    )
    fourier_potential = FourierVoxelGridPotential.from_real_voxel_grid(
        potential_as_real_voxel_grid, voxel_size
    )
    # Since Voxelgrid is in Frequency space by default, we have to first
    # transform back into real space.
    fvg_real = ifftn(jnp.fft.ifftshift(fourier_potential.fourier_voxel_grid)).real

    vg = RealVoxelGridPotential.from_real_voxel_grid(
        potential_as_real_voxel_grid, voxel_size
    )

    np.testing.assert_allclose(fvg_real, vg.real_voxel_grid, atol=1e-12)


def test_downsampled_voxel_potential_agreement(sample_pdb_path):
    """Integration test ensuring that rasterized voxel grids roughly
    agree with downsampled versions.
    """
    # Parameters for rasterization
    shape = (128, 128, 128)
    voxel_size = 0.25
    # Downsampling parameters
    downsampling_factor = 2
    downsampled_shape = (
        int(shape[0] / downsampling_factor),
        int(shape[1] / downsampling_factor),
        int(shape[2] / downsampling_factor),
    )
    downsampled_voxel_size = voxel_size * downsampling_factor
    # Load the PDB file
    atom_positions, atom_elements = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        loads_b_factors=False,
        selection_string="not element H",
    )
    # Load atomistic potential
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_elements, read_peng_element_scattering_factor_parameter_table()
    )
    atomic_potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
    )
    # Build the grids
    low_resolution_potential_grid = atomic_potential.as_real_voxel_grid(
        downsampled_shape, downsampled_voxel_size
    )
    high_resolution_potential_grid = atomic_potential.as_real_voxel_grid(
        shape, voxel_size
    )
    downsampled_potential_grid = downsample_with_fourier_cropping(
        high_resolution_potential_grid, downsampling_factor
    )

    assert low_resolution_potential_grid.shape == downsampled_potential_grid.shape


@pytest.mark.parametrize(
    "batch_size_for_z_planes,n_batches_of_atoms",
    ((1, 1), (2, 1), (3, 1), (1, 2), (1, 3), (2, 2)),
)
def test_z_plane_batched_vs_non_batched_loop_agreement(
    sample_pdb_path, batch_size_for_z_planes, n_batches_of_atoms
):
    shape = (128, 128, 128)
    voxel_size = 0.5

    # Load the PDB file
    atom_positions, atom_elements = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        loads_b_factors=False,
        selection_string="not element H",
    )
    # Load atomistic potential
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_elements, read_peng_element_scattering_factor_parameter_table()
    )
    atomic_potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
    )
    # Build the grid
    voxels = atomic_potential.as_real_voxel_grid(shape, voxel_size)
    voxels_with_batching = atomic_potential.as_real_voxel_grid(
        shape,
        voxel_size,
        batch_size_for_z_planes=batch_size_for_z_planes,
        n_batches_of_atoms=n_batches_of_atoms,
    )
    np.testing.assert_allclose(voxels, voxels_with_batching)


@pytest.mark.parametrize("shape", ((128, 127, 126),))
def test_compute_rectangular_voxel_grid(sample_pdb_path, shape):
    voxel_size = 0.5

    # Load the PDB file
    atom_positions, atom_elements = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        loads_b_factors=False,
        selection_string="not element H",
    )
    # Load atomistic potential
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_elements, read_peng_element_scattering_factor_parameter_table()
    )
    atomic_potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
    )
    # Build the grid
    voxels = atomic_potential.as_real_voxel_grid(shape, voxel_size)
    assert voxels.shape == shape


class TestBuildRealSpaceVoxelsFromAtoms:
    @pytest.mark.parametrize("largest_atom", range(0, 3))
    def test_maxima_are_in_right_positions(self, toy_gaussian_cloud, largest_atom):
        """
        Test that the maxima of the potential are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud
        ff_a = ff_a.at[largest_atom].add(1.0)

        # Build the potential
        atomic_potential = GaussianMixtureAtomicPotential(
            atom_positions, ff_a, ff_b / (8 * jnp.pi**2)
        )
        real_voxel_grid = atomic_potential.as_real_voxel_grid(
            n_voxels_per_side, voxel_size
        )
        coordinate_grid = make_coordinate_grid(n_voxels_per_side, voxel_size)

        # Find the maximum
        maximum_index = jnp.argmax(real_voxel_grid)
        maximum_position = coordinate_grid.reshape(-1, 3)[maximum_index]

        # Check that the maximum is in the correct position
        assert jnp.allclose(maximum_position, atom_positions[largest_atom])

    def test_integral_is_correct(self, toy_gaussian_cloud):
        """
        Test that the maxima of the potential are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud

        # Build the potential
        atomic_potential = GaussianMixtureAtomicPotential(
            atom_positions, ff_a, ff_b / (8 * jnp.pi**2)
        )
        real_voxel_grid = atomic_potential.as_real_voxel_grid(
            n_voxels_per_side, voxel_size
        )

        integral = jnp.sum(real_voxel_grid) * voxel_size**3
        assert jnp.isclose(integral, jnp.sum(4 * jnp.pi * ff_a))

    # TODO: Can we parallelize this test? Runs a little too restrictively slow on the GPU
    # def test_fourier_transform(self, toy_gaussian_cloud):
    #     import itertools
    #     (
    #         atom_positions,
    #         ff_a,
    #         ff_b,
    #         n_voxels_per_side,
    #         voxel_size,
    #     ) = toy_gaussian_cloud
    #     coordinate_grid = make_coordinate_grid(n_voxels_per_side, voxel_size)
    #     # Build the potential
    #     atomic_potential = GaussianMixtureAtomicPotential(
    #         atom_positions,
    #         ff_a,
    #         ff_b / (8 * jnp.pi**2)
    #     )
    #     real_voxel_grid = atomic_potential.as_real_voxel_grid(coordinate_grid)
    #     fourier_potential = FourierVoxelGridPotential.from_real_voxel_grid(
    #         real_voxel_grid, voxel_size
    #     )
    #     fourier_voxel_grid = fourier_potential.fourier_voxel_grid
    #     bin_size = 1 / (voxel_size) / n_voxels_per_side[0]
    #     num_atoms = atom_positions.shape[0]

    #     translational_phase_factor = lambda i, j, k, atom_position: jnp.exp(
    #         -1j
    #         * 2
    #         * jnp.pi
    #         * jnp.dot(atom_position, jnp.array([i - 64, j - 64, k - 64]))
    #         / (n_voxels_per_side[0] * voxel_size)
    #     )
    #     # Verify generated fourier_voxel_grid agrees with scattering equation in Peng.
    #     # Check up to 1/4 Nyquist Frequency in each axis.
    #     frequency_array = np.fromiter(
    #         itertools.product(np.arange(64, 80), repeat=3), "i,i,i"
    #     ).view(("i", 3))
    #     for frequency in frequency_array:
    #         [i, j, k] = frequency
    #         predicted_value = 0
    #         frequency_magnitude = np.sqrt((64 - i) ** 2 + (64 - j) ** 2 + (64 - k) ** 2)
    #         for atom_index in range(0, num_atoms):
    #             atom_contribution = jnp.sum(
    #                 ff_a[atom_index]
    #                 * 4
    #                 * jnp.pi
    #                 * (voxel_size) ** -3
    #                 * jnp.exp(
    #                     -ff_b[atom_index]
    #                     * ((1 / 2) * bin_size * frequency_magnitude) ** 2
    #                 )
    #                 * translational_phase_factor(i, j, k, atom_positions[atom_index])
    #             )

    #             predicted_value = predicted_value + atom_contribution

    #         fourier_grid_value = fourier_voxel_grid[k][j][i]
    #         assert jnp.isclose((predicted_value), (fourier_grid_value), rtol=1e-4)


class TestBuildVoxelsFromTrajectories:
    def test_indexing_matches_individual_calls(self, toy_gaussian_cloud):
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud
        second_set_of_positions = atom_positions + 1.0
        traj = jnp.stack([atom_positions, second_set_of_positions], axis=0)

        make_voxel_grids = jax.vmap(
            lambda pos, ff_a, ff_b: GaussianMixtureAtomicPotential(
                pos, ff_a, ff_b / (8 * jnp.pi**2)
            ).as_real_voxel_grid(n_voxels_per_side, voxel_size),
            in_axes=[0, None, None],
        )
        traj_voxels = make_voxel_grids(traj, ff_a, ff_b)

        voxel1 = GaussianMixtureAtomicPotential(
            atom_positions, ff_a, ff_b / (8 * jnp.pi**2)
        ).as_real_voxel_grid(n_voxels_per_side, voxel_size)

        voxel2 = GaussianMixtureAtomicPotential(
            second_set_of_positions, ff_a, ff_b / (8 * jnp.pi**2)
        ).as_real_voxel_grid(n_voxels_per_side, voxel_size)

        np.testing.assert_allclose(traj_voxels[0], voxel1, atol=1e-12)
        np.testing.assert_allclose(traj_voxels[1], voxel2, atol=1e-12)



import os

import numpy as np
import pytest
from jax import config
from jaxtyping import install_import_hook


with install_import_hook("cryojax", "typeguard.typechecked"):
    from cryojax.io import read_atoms_from_pdb

config.update("jax_enable_x64", True)


@pytest.fixture
def pdb_multiple_structures_path():
    return os.path.join(os.path.dirname(__file__), "data", "1uao_assembly.pdb")


def test_read_structure(sample_pdb_path):
    atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="protein and not element H",
        loads_b_factors=True,
    )

    assert atom_positions.ndim == 2
    assert atom_identities.shape == b_factors.shape
    assert atom_positions.shape[0] == atom_identities.shape[0]

    assert atom_positions.shape[1] == 3
    assert atom_positions.shape[0] == 77


def test_read_structure_no_b_factors(sample_pdb_path):
    atom_positions, atom_identities = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="protein and not element H",
        loads_b_factors=False,
    )

    assert atom_positions.ndim == 2
    assert atom_positions.shape[0] == atom_identities.shape[0]

    assert atom_positions.shape[1] == 3
    assert atom_positions.shape[0] == 77


def test_read_pdb_multiple_structures(pdb_multiple_structures_path):
    atom_positions, atom_identities = read_atoms_from_pdb(
        pdb_multiple_structures_path,
        center=True,
        loads_b_factors=False,
        selection_string="all",
    )
    assert atom_positions.ndim == 3
    assert atom_positions.shape[0] == 10
    assert atom_positions.shape[1] == atom_identities.shape[0]
    assert atom_positions.shape[2] == 3
    assert atom_positions.shape[1] == 138


def test_read_pdb_at_structure(pdb_multiple_structures_path):
    atom_positions, atom_identities = read_atoms_from_pdb(
        pdb_multiple_structures_path,
        center=True,
        loads_b_factors=False,
        selection_string="all",
        model_index=1,
    )

    assert atom_positions.ndim == 2
    assert atom_positions.shape[0] == atom_identities.shape[0]
    assert atom_positions.shape[1] == 3
    assert atom_positions.shape[0] == 138


# def test_read_full_assembly_pdb(sample_pdb_path_assembly):
#     atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
#         sample_pdb_path_assembly,
#         center=True,
#         selection_string="all",
#         is_assembly=True,
#         i_model=None,
#         loads_b_factors=True,
#     )

#     assert atom_positions.ndim == 2
#     assert atom_identities.shape == b_factors.shape
#     assert atom_positions.shape[0] == atom_identities.shape[0]

#     assert atom_positions.shape[1] == 3
#     assert atom_positions.shape[0] == 1380


# def test_read_cif(sample_cif_path):
#     atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
#         sample_cif_path,
#         center=True,
#         selection_string="all",
#         is_assembly=False,
#         i_model=None,
#         loads_b_factors=True,
#     )

#     assert atom_positions.ndim == 2
#     assert atom_identities.shape == b_factors.shape
#     assert atom_positions.shape[0] == atom_identities.shape[0]

#     assert atom_positions.shape[1] == 3
#     assert atom_positions.shape[0] == 3222


def test_read_from_url(sample_pdb_url):
    atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
        sample_pdb_url,
        center=True,
        selection_string="all",
        loads_b_factors=True,
    )

    assert atom_positions.ndim == 2
    assert atom_identities.shape == b_factors.shape
    assert atom_positions.shape[0] == atom_identities.shape[0]

    assert atom_positions.shape[1] == 3
    assert atom_positions.shape[0] == 1973


def test_center_waterbox(sample_waterbox_pdb):
    atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
        sample_waterbox_pdb,
        center=True,
        selection_string="all",
        loads_b_factors=True,
    )

    assert not np.isnan(atom_positions).any(), "Centering resulted in positions with NaNs"



"""
Using non-uniform FFTs for computing volume projections.
"""

import math
from typing import ClassVar, Optional
from typing_extensions import override

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ....ndimage import convert_fftn_to_rfftn, irfftn
from ..._config import AbstractConfig
from ..._potential_representation import RealVoxelCloudPotential, RealVoxelGridPotential
from .base_direct_integrator import AbstractDirectVoxelIntegrator


class NufftProjection(
    AbstractDirectVoxelIntegrator[RealVoxelGridPotential | RealVoxelCloudPotential],
    strict=True,
):
    """Integrate points onto the exit plane using non-uniform FFTs."""

    pixel_rescaling_mode: Optional[str]
    eps: float

    is_projection_approximation: ClassVar[bool] = True
    requires_inverse_rotation: ClassVar[bool] = False

    def __init__(self, *, pixel_rescaling_mode: Optional[str] = None, eps: float = 1e-6):
        """**Arguments:**

        - `pixel_rescaling_mode`: Method for interpolating the final image to
                                    the `AbstractConfig` pixel size. See
                                    `cryojax.image.rescale_pixel_size` for documentation.
        - `eps`: See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
                 for documentation.
        """
        self.pixel_rescaling_mode = pixel_rescaling_mode
        self.eps = eps

    def project_voxel_cloud_with_nufft(
        self,
        weights: Float[Array, " size"],
        coordinate_list_in_angstroms: Float[Array, "size 2"] | Float[Array, "size 3"],
        shape: tuple[int, int],
    ) -> Complex[Array, "{shape[0]} {shape[1]//2+1}"]:
        """Project and interpolate 3D volume point cloud
        onto imaging plane using a non-uniform FFT.

        **Arguments:**

        - `weights`:
            Density point cloud.
        - `coordinate_list_in_angstroms`:
            Coordinate system of point cloud.
        - `shape`:
            Shape of the real-space imaging plane in pixels.

        **Returns:**

        The fourier-space projection of the density point cloud defined by `weights` and
        `coordinate_list_in_angstroms`.
        """
        return _project_with_nufft(weights, coordinate_list_in_angstroms, shape, self.eps)

    @override
    def integrate(
        self,
        potential: RealVoxelGridPotential | RealVoxelCloudPotential,
        config: AbstractConfig,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim//2+1}",
        ]
        | Float[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
    ):
        """Compute the integrated scattering potential at the `AbstractConfig` settings
        of a voxel-based representation in real-space, using non-uniform FFTs.

        **Arguments:**

        - `potential`: The scattering potential representation.
        - `config`: The configuration of the resulting image.

        **Returns:**

        The projection integral of the `potential` in fourier space, at the
        `config.padded_shape` and the `config.pixel_size`.
        """
        if isinstance(potential, RealVoxelGridPotential):
            shape = potential.shape
            fourier_in_plane_potential = self.project_voxel_cloud_with_nufft(
                potential.real_voxel_grid.ravel(),
                potential.coordinate_grid_in_pixels.reshape((math.prod(shape), 3)),
                config.padded_shape,
            )
        elif isinstance(potential, RealVoxelCloudPotential):
            fourier_in_plane_potential = self.project_voxel_cloud_with_nufft(
                potential.voxel_weights,
                potential.coordinate_list_in_pixels,
                config.padded_shape,
            )
        else:
            raise ValueError(
                "Supported types for `potential` are `RealVoxelGridPotential` and "
                "`RealVoxelCloudPotential`."
            )
        # Scale by voxel size to convert from projection to integral
        fourier_in_plane_potential *= potential.voxel_size
        # Re-scale to correct pixel size if its different
        fourier_in_plane_potential = self._maybe_rescale_pixel_size(
            fourier_in_plane_potential, potential, config, input_is_rfft=True
        )
        return (
            irfftn(fourier_in_plane_potential, s=config.padded_shape)
            if outputs_real_space
            else fourier_in_plane_potential
        )


def _project_with_nufft(weights, coordinate_list, shape, eps=1e-6):
    from jax_finufft import nufft1

    weights, coordinate_list = (
        jnp.asarray(weights).astype(complex),
        jnp.asarray(coordinate_list),
    )
    # Get x and y coordinates
    coordinates_xy = coordinate_list[:, :2]
    # Normalize coordinates betweeen -pi and pi
    ny, nx = shape
    box_xy = jnp.asarray((nx, ny), dtype=float)
    coordinates_periodic = 2 * jnp.pi * coordinates_xy / box_xy
    # Unpack and compute
    x, y = coordinates_periodic[:, 0], coordinates_periodic[:, 1]
    fourier_projection = nufft1(shape, weights, y, x, eps=eps, iflag=-1)
    # Shift zero frequency component to corner
    fourier_projection = jnp.fft.ifftshift(fourier_projection)
    # Convert to rfftn output
    return convert_fftn_to_rfftn(fourier_projection, mode="real")



import jax.numpy as jnp
import numpy as np
import pytest
from jax import config
from jaxtyping import install_import_hook


with install_import_hook("cryojax", "typeguard.typechecked"):
    from cryojax.constants import (
        get_tabulated_scattering_factor_parameters,
        read_peng_element_scattering_factor_parameter_table,
    )
    from cryojax.coordinates import make_coordinate_grid
    from cryojax.io import read_atoms_from_pdb
    from cryojax.ndimage import irfftn
    from cryojax.simulator import (
        BasicConfig,
        GaussianMixtureAtomicPotential,
        GaussianMixtureProjection,
        PengAtomicPotential,
    )


config.update("jax_enable_x64", True)


@pytest.fixture
def toy_gaussian_cloud():
    atom_positions = jnp.array(
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ]
    )
    num_atoms = atom_positions.shape[0]
    ff_a = jnp.array(
        num_atoms
        * [
            [1.0, 0.5],
        ]
    )

    ff_b = jnp.array(
        num_atoms
        * [
            [0.3, 0.2],
        ]
    )

    n_voxels_per_side = (128, 128, 128)
    voxel_size = 0.05
    return (atom_positions, ff_a, ff_b, n_voxels_per_side, voxel_size)


@pytest.mark.parametrize("shape", ((64, 64), (63, 63), (63, 64), (64, 63)))
def test_atom_integrator_shape(sample_pdb_path, shape):
    atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="not element H",
        loads_b_factors=True,
    )
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_identities, read_peng_element_scattering_factor_parameter_table()
    )
    atom_potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
        b_factors=b_factors,
    )
    pixel_size = 0.5

    integrator = GaussianMixtureProjection(upsampling_factor=2)
    # # ... and the configuration of the imaging instrument
    config = BasicConfig(
        shape=shape,
        pixel_size=pixel_size,
        voltage_in_kilovolts=300.0,
    )
    # ... compute the integrated volumetric_potential
    fourier_integrated_potential = integrator.integrate(
        atom_potential, config, outputs_real_space=False
    )

    assert fourier_integrated_potential.shape == (shape[0], shape[1] // 2 + 1)


def test_downsampled_gmm_potential_agreement(sample_pdb_path):
    """Integration test ensuring that rasterized voxel grids roughly
    agree with downsampled versions.
    """
    atom_positions, atom_identities = read_atoms_from_pdb(
        sample_pdb_path,
        loads_b_factors=False,
        center=True,
        selection_string="not element H",
    )
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_identities, read_peng_element_scattering_factor_parameter_table()
    )
    atom_potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
    )

    # Parameters for rasterization
    shape = (128, 128)
    pixel_size = 0.25

    # Downsampling parameters
    downsampling_factor = 2
    downsampled_shape = (
        int(shape[0] / downsampling_factor),
        int(shape[1] / downsampling_factor),
    )
    downsampled_pixel_size = pixel_size * downsampling_factor

    integrator_int_hires = GaussianMixtureProjection(
        upsampling_factor=downsampling_factor
    )
    integrator_int_lowres = GaussianMixtureProjection(upsampling_factor=1)
    # ... and the configuration of the imaging instrument
    config = BasicConfig(
        shape=downsampled_shape,
        pixel_size=downsampled_pixel_size,
        voltage_in_kilovolts=300.0,
    )
    # ... compute the integrated volumetric_potential
    image_from_hires = integrator_int_hires.integrate(atom_potential, config)
    image_lowres = integrator_int_lowres.integrate(atom_potential, config)

    assert image_from_hires.shape == image_lowres.shape


def test_peng_vs_gmm_agreement(sample_pdb_path):
    """Integration test ensuring that Peng Potential and GMM potential agree when
    gaussians are identical"""

    # Load atoms and build potentials
    atom_positions, atom_identities = read_atoms_from_pdb(
        sample_pdb_path,
        loads_b_factors=False,
        center=True,
        selection_string="not element H",
    )
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_identities, read_peng_element_scattering_factor_parameter_table()
    )
    atom_potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
    )

    gaussian_variances = atom_potential.scattering_factor_b / (8.0 * jnp.pi**2)
    gaussian_amplitudes = atom_potential.scattering_factor_a

    gmm_potential = GaussianMixtureAtomicPotential(
        atom_positions, gaussian_amplitudes, gaussian_variances
    )

    # Create instrument configuration
    shape = (64, 64)
    pixel_size = 0.5
    config = BasicConfig(
        shape=shape,
        pixel_size=pixel_size,
        voltage_in_kilovolts=300.0,
    )

    # Compute projections
    integrator = GaussianMixtureProjection(upsampling_factor=1)
    projection_gmm = integrator.integrate(gmm_potential, config)
    projection_peng = integrator.integrate(atom_potential, config)

    np.testing.assert_allclose(projection_gmm, projection_peng)


class TestBuildRealSpaceVoxelsFromAtoms:
    @pytest.mark.parametrize("largest_atom", range(0, 3))
    def test_maxima_are_in_right_positions(self, toy_gaussian_cloud, largest_atom):
        """
        Test that the maxima of the potential are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud

        n_pixels_per_side = n_voxels_per_side[:2]
        ff_a = ff_a.at[largest_atom].add(1.0)
        coordinate_grid = make_coordinate_grid(n_pixels_per_side, voxel_size)

        # Build the potential
        atomic_potential = GaussianMixtureAtomicPotential(
            atom_positions, ff_a, ff_b / (8.0 * jnp.pi**2)
        )
        config = BasicConfig(
            shape=n_pixels_per_side,
            pixel_size=voxel_size,
            voltage_in_kilovolts=300.0,
        )
        # Build the potential integrators
        integrator = GaussianMixtureProjection()
        # Compute projections
        projection = integrator.integrate(atomic_potential, config)
        projection = irfftn(projection)

        # Find the maximum
        maximum_index = jnp.argmax(projection)
        maximum_position = coordinate_grid.reshape(-1, 2)[maximum_index]

        # Check that the maximum is in the correct position
        assert jnp.allclose(maximum_position, atom_positions[largest_atom][:2])

    def test_integral_is_correct(self, toy_gaussian_cloud):
        """
        Test that the maxima of the potential are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud

        n_pixels_per_side = n_voxels_per_side[:2]
        # Build the potential
        atomic_potential = GaussianMixtureAtomicPotential(
            atom_positions, ff_a, ff_b / (8.0 * jnp.pi**2)
        )
        config = BasicConfig(
            shape=n_pixels_per_side,
            pixel_size=voxel_size,
            voltage_in_kilovolts=300.0,
        )
        # Build the potential integrators
        integrator = GaussianMixtureProjection()
        # Compute projections
        projection = integrator.integrate(atomic_potential, config)
        projection = irfftn(projection)

        integral = jnp.sum(projection) * voxel_size**2
        assert jnp.isclose(integral, jnp.sum(4 * jnp.pi * ff_a))



from .base_ensemble import (
    AbstractStructuralEnsemble as AbstractStructuralEnsemble,
)
from .base_structure import (
    AbstractStructure as AbstractStructure,
    BasicStructure as BasicStructure,
)
from .discrete_ensemble import (
    DiscreteStructuralEnsemble as DiscreteStructuralEnsemble,
)



"""
Read atomic information from a PDB file. Functions and objects are
adapted from `mdtraj`.
"""

import bz2
import gzip
import io
import os
import pathlib
from copy import copy
from io import StringIO
from typing import Literal, Optional, TypedDict, cast, overload
from urllib.parse import urlparse, uses_netloc, uses_params, uses_relative
from urllib.request import urlopen
from xml.etree import ElementTree

import jax
import mdtraj
import numpy as np
from jaxtyping import Float, Int
from mdtraj.core import element as elem
from mdtraj.core.topology import Topology
from mdtraj.formats.pdb.pdbstructure import PdbStructure


_VALID_URLS = set(uses_relative + uses_netloc + uses_params)
_VALID_URLS.discard("")


@overload
def read_atoms_from_pdb(
    filename_or_url: str | pathlib.Path,
    loads_b_factors: Literal[False],
    center: bool = True,
    *,
    selection_string: str = "all",
    model_index: Optional[int] = None,
    standardizes_names: bool = True,
    topology: Optional[mdtraj.Topology] = None,
) -> tuple[Float[np.ndarray, "... n_atoms 3"], Int[np.ndarray, " n_atoms"]]: ...


@overload
def read_atoms_from_pdb(
    filename_or_url: str | pathlib.Path,
    loads_b_factors: Literal[True],
    center: bool = True,
    *,
    selection_string: str = "all",
    model_index: Optional[int] = None,
    standardizes_names: bool = True,
    topology: Optional[mdtraj.Topology] = None,
) -> tuple[
    Float[np.ndarray, "... n_atoms 3"],
    Int[np.ndarray, " n_atoms"],
    Float[np.ndarray, " n_atoms"],
]: ...


@overload
def read_atoms_from_pdb(
    filename_or_url: str | pathlib.Path,
    loads_b_factors: bool = False,
    center: bool = True,
    *,
    selection_string: str,
    model_index: Optional[int] = None,
    standardizes_names: bool = True,
    topology: Optional[mdtraj.Topology] = None,
): ...


def read_atoms_from_pdb(
    filename_or_url: str | pathlib.Path,
    loads_b_factors: bool = False,
    center: bool = True,
    *,
    selection_string: str = "all",
    model_index: Optional[int] = None,
    standardizes_names: bool = True,
    topology: Optional[mdtraj.Topology] = None,
) -> (
    tuple[Float[np.ndarray, "... n_atoms 3"], Int[np.ndarray, " n_atoms"]]
    | tuple[
        Float[np.ndarray, "... n_atoms 3"],
        Int[np.ndarray, " n_atoms"],
        Float[np.ndarray, " n_atoms"],
    ]
):
    """Read atomic information from a PDB file. This object
    wraps the `cryojax.io.AtomicModelReader` class into a function
    interface to accomodate most use cases in cryo-EM.

    **Arguments:**

    - `filename_or_url`:
        The name of the PDB/mmCIF file to open. Can be a URL.
    - `center`:
        If `True`, center the model so that its center of mass coincides
        with the origin.
    - `loads_b_factors`:
        If `True`, return the B-factors of the atoms.
    - `selection_string`:
        A selection string in `mdtraj`'s format. See `mdtraj` for documentation.
    - `model_index`:
        An optional index for grabbing a particular model stored in the PDB. If `None`,
        grab all models, where `atom_positions` has a leading dimension for the model.
    - `standardizes_names`:
        If `True`, non-standard atom names and residue names are standardized to conform
        with the current PDB format version. If set to `False`, this step is skipped.
    - `topology`:
        If you give a topology as input, the topology won't be parsed from the pdb file
        it saves time if you have to parse a big number of files

    **Returns:**

    A tuple whose first element is a `numpy` array of coordinates containing
    atomic positions, and whose second element is an array of atomic element
    numbers. To be clear,

    ```python
    atom_positons, atom_identities = read_atoms_from_pdb(...)
    ```

    !!! info

        If your PDB has multiple models, `atom_positions` by
        default with a leading dimension that indexes each model.
        On the other hand, `atom_identities` (and `b_factors`, if loaded)
        do not have this leading dimension and are constant across
        models.
    """
    # Load all atoms
    with AtomicModelFile(filename_or_url) as pdb_file:
        # Read file
        atom_info, topology = pdb_file.read_atoms(
            standardizes_names=standardizes_names,
            topology=topology,
            model_index=model_index,
            loads_masses=center,
            loads_b_factors=loads_b_factors,
        )
    # Filter atoms and grab positions and identities
    selected_indices = topology.select(selection_string)
    atom_positions = atom_info["positions"][:, selected_indices]
    atom_properties = jax.tree.map(
        lambda arr: arr[selected_indices], atom_info["properties"]
    )
    atom_identities = atom_properties["identities"]
    # Center by mass
    if center:
        atom_masses = cast(np.ndarray, atom_properties["masses"])
        atom_positions = _center_atom_coordinates(atom_positions, atom_masses)
    # Return, optionality with b-factors and without a leading dimension for the
    # positions if there is only one structure
    if atom_positions.shape[0] == 1:
        atom_positions = np.squeeze(atom_positions, axis=0)
    if loads_b_factors:
        b_factors = cast(np.ndarray, atom_properties["b_factors"])
        return atom_positions, atom_identities, b_factors
    else:
        return atom_positions, atom_identities


def _center_atom_coordinates(atom_positions, atom_masses):
    com_position = np.transpose(atom_positions, axes=[0, 2, 1]).dot(
        atom_masses / atom_masses.sum()
    )
    return atom_positions - com_position[:, None, :]


class AtomProperties(TypedDict):
    """A struct for the info of individual atoms.

    **Attributes:**

    - `identities`: The atomic numbers of all of the atoms.
    - `masses`: The mass of each atom.
    - `b_factors`: The B-factors of all of the atoms.
    """

    identities: Int[np.ndarray, " N"]
    masses: Optional[Float[np.ndarray, " N"]]
    b_factors: Optional[Float[np.ndarray, " N"]]


class AtomicModelInfo(TypedDict):
    """A struct for the info of individual atoms.

    **Keys:**

    - `positions`:
        The cartesian coordinates of all of the atoms.
    - `properties`:
        The static properties of the individual atom, i.e.
        that do not change between frames.
    """

    positions: Float[np.ndarray, "M N 3"]
    properties: AtomProperties


class AtomicModelFile:
    """A PDB file loader that loads the necessary information for
    cryo-EM. This object is based on the `PDBTrajectoryFile`
    from `mdtraj`.
    """

    def __init__(self, filename_or_url: str | pathlib.Path):
        """**Arguments:**

        - `filename_or_url`:
            The name of the PDB file to open. Can be a URL.
        """
        # Set field that says we are reading the file
        self._is_open = True
        # Setup I/O
        if _is_url(filename_or_url):
            filename_or_url = str(filename_or_url)
            file_obj = urlopen(filename_or_url)
            if filename_or_url.lower().endswith(".gz"):
                file_obj = gzip.GzipFile(fileobj=file_obj)
            self._file = StringIO(file_obj.read().decode("utf-8"))

        else:
            filename_or_url = pathlib.Path(filename_or_url)
            _validate_pdb_file(filename_or_url)
            self._file = _open_maybe_zipped(filename_or_url, "r")
        # Create the PDB structure via `mdtraj`
        self._pdb_structure = PdbStructure(self._file, load_all_models=True)

    @property
    def pdb_structure(self) -> PdbStructure:
        return self._pdb_structure

    def read_atoms(
        self,
        *,
        standardizes_names: bool = True,
        topology: Optional[Topology] = None,
        model_index: Optional[int] = None,
        loads_b_factors: bool = True,
        loads_masses: bool = True,
    ) -> tuple[AtomicModelInfo, Topology]:
        """Load properties from the PDB reader.

        **Arguments:**

        - `standardizes_names`:
            If `True`, non-standard atom names and residue names are standardized to
            conform with the current PDB format version. If `False`, this step is skipped.
        - `topology`:
            If the `topology` is passed as input, it won't be parsed from the PDB file.
            This saves time if you have to parse a big number of files.

        **Returns:**

        A tuple of the `AtomicModelInfo` dataclass and the `mdtraj.Topology`.
        """
        atom_info, topology = _load_atom_info(
            self._pdb_structure,
            topology,
            model_index,
            standardizes_names,
            loads_b_factors,
            loads_masses,
        )

        return atom_info, topology

    @property
    def is_open(self):
        return self._is_open

    def close(self):
        """Close the PDB file"""
        if self._is_open:
            if hasattr(self, "_file"):
                self._file.close()
        self._is_open = False

    def __del__(self):
        self.close()
        del self

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()


def _make_topology(
    pdb: PdbStructure,
    atom_positions: list,
    standardizes_names: bool,
    model_index: Optional[int],
) -> Topology:
    topology = Topology()
    if standardizes_names:
        residue_name_replacements, atom_name_replacements = (
            _load_name_replacement_tables()
        )
    else:
        residue_name_replacements, atom_name_replacements = {}, {}
    atom_by_number = {}
    if model_index is None:
        model_index = 0
    model = pdb.models_by_number[model_index]
    for chain in model.iter_chains():
        c = topology.add_chain(chain.chain_id)
        for residue in chain.iter_residues():
            residue_name = residue.get_name()
            if residue_name in residue_name_replacements and standardizes_names:
                residue_name = residue_name_replacements[residue_name]
            r = topology.add_residue(residue_name, c, residue.number, residue.segment_id)
            if residue_name in atom_name_replacements and standardizes_names:
                atom_replacements = atom_name_replacements[residue_name]
            else:
                atom_replacements = {}
            for atom in residue.atoms:
                atom_name = atom.get_name()
                if atom_name in atom_replacements:
                    atom_name = atom_replacements[atom_name]
                atom_name = atom_name.strip()
                element = atom.element
                if element is None:
                    element = _guess_element(atom_name, residue.name, len(residue))

                new_atom = topology.add_atom(
                    atom_name,
                    element,
                    r,
                    serial=atom.serial_number,
                    formal_charge=atom.formal_charge,
                )
                atom_by_number[atom.serial_number] = new_atom

    topology.create_standard_bonds()
    topology.create_disulfide_bonds(atom_positions[0])

    return topology


def _load_atom_info(
    pdb: PdbStructure,
    topology: Optional[Topology],
    model_index: Optional[int],
    standardizes_names: bool,
    loads_b_factors: bool,
    loads_masses: bool,
):
    # Load atom info
    if model_index is None:
        # ... with multiple models
        n_models = len(pdb.model_numbers())
        temp = dict(
            positions=[*([] for _ in range(n_models))],
            identities=[],
            b_factors=[],
            masses=[],
        )
        for index, model_index in enumerate(pdb.model_numbers()):  # type: ignore
            model = pdb.models_by_number[model_index]
            for chain in model.iter_chains():
                for residue in chain.iter_residues():
                    for atom in residue.atoms:
                        # ... make sure this is read in angstroms?
                        temp["positions"][index].append(atom.get_position())
                        if index == 0:
                            # Assume atom properties don't change between models
                            temp["identities"].append(atom.element.atomic_number)
                            if loads_masses:
                                temp["masses"].append(atom.element.mass)
                            if loads_b_factors:
                                temp["b_factors"].append(atom.get_temperature_factor())
    else:
        # ... with a model at one index
        n_models = 1
        temp = dict(positions=[[]], identities=[], b_factors=[], masses=[])
        try:
            model = pdb.models_by_number[model_index]
        except Exception as err:
            raise ValueError(
                "Caught exception indexing atomic model with "
                f"index {model_index}. Found that the PDB "
                f"contained model numbers {pdb.model_numbers()} "
                f"available for indexing. Traceback was:\n{err}"
            )
        for chain in model.iter_chains():
            for residue in chain.iter_residues():
                for atom in residue.atoms:
                    # ... make sure this is read in angstroms?
                    temp["positions"][0].append(atom.get_position())
                    # Assume atom properties don't change between models
                    temp["identities"].append(atom.element.atomic_number)
                    if loads_masses:
                        temp["masses"].append(atom.element.mass)
                    if loads_b_factors:
                        temp["b_factors"].append(atom.get_temperature_factor())

    # Load the topology if None is given
    if topology is None:
        topology = _make_topology(pdb, temp["positions"], standardizes_names, model_index)

    # Gather atom info and return
    properties = AtomProperties(
        identities=np.asarray(temp["identities"], dtype=int),
        b_factors=(
            np.asarray(temp["b_factors"], dtype=float) if loads_b_factors else None
        ),
        masses=(np.asarray(temp["masses"], dtype=float) if loads_masses else None),
    )
    atom_info = AtomicModelInfo(
        positions=np.asarray(temp["positions"], dtype=float),
        properties=properties,
    )

    return atom_info, topology


def _guess_element(atom_name, residue_name, residue_length):
    """Try to guess the element name.
    Closely follows `mdtraj.formats.pdb.PDBTrajectoryFile._guess_element`."""
    upper = atom_name.upper()
    if upper.startswith("CL"):
        element = elem.chlorine
    elif upper.startswith("NA"):
        element = elem.sodium
    elif upper.startswith("MG"):
        element = elem.magnesium
    elif upper.startswith("BE"):
        element = elem.beryllium
    elif upper.startswith("LI"):
        element = elem.lithium
    elif upper.startswith("K"):
        element = elem.potassium
    elif upper.startswith("ZN"):
        element = elem.zinc
    elif residue_length == 1 and upper.startswith("CA"):
        element = elem.calcium

    # TJL has edited this. There are a few issues here. First,
    # parsing for the element is non-trivial, so I do my best
    # below. Second, there is additional parsing code in
    # pdbstructure.py, and I am unsure why it doesn't get used
    # here...
    elif residue_length > 1 and upper.startswith("CE"):
        element = elem.carbon  # (probably) not Celenium...
    elif residue_length > 1 and upper.startswith("CD"):
        element = elem.carbon  # (probably) not Cadmium...
    elif residue_name in ["TRP", "ARG", "GLN", "HIS"] and upper.startswith("NE"):
        element = elem.nitrogen  # (probably) not Neon...
    elif residue_name in ["ASN"] and upper.startswith("ND"):
        element = elem.nitrogen  # (probably) not ND...
    elif residue_name == "CYS" and upper.startswith("SG"):
        element = elem.sulfur  # (probably) not SG...
    else:
        try:
            element = elem.get_by_symbol(atom_name[0])
        except KeyError:
            try:
                symbol = (
                    atom_name[0:2].strip().rstrip("AB0123456789").lstrip("0123456789")
                )
                element = elem.get_by_symbol(symbol)
            except KeyError:
                element = None

    return element


def _load_name_replacement_tables():
    """Load the list of atom and residue name replacements.
    Closely follows `mdtraj.formats.pdb.PDBTrajectoryFile._loadNameReplacementTables`.
    """
    tree = ElementTree.parse(
        os.path.join(os.path.dirname(__file__), "pdbNames.xml"),
    )
    # Residue and atom replacements
    residue_name_replacements = {}
    atom_name_replacements = {}
    # ... containers for residues
    all_residues, protein_residues, nucleic_acid_residues = {}, {}, {}
    for residue in tree.getroot().findall("Residue"):
        name = residue.attrib["name"]
        if name == "All":
            _parse_residue(residue, all_residues)
        elif name == "Protein":
            _parse_residue(residue, protein_residues)
        elif name == "Nucleic":
            _parse_residue(residue, nucleic_acid_residues)
    for atom in all_residues:
        protein_residues[atom] = all_residues[atom]
        nucleic_acid_residues[atom] = all_residues[atom]
    for residue in tree.getroot().findall("Residue"):
        name = residue.attrib["name"]
        for id in residue.attrib:
            if id == "name" or id.startswith("alt"):
                residue_name_replacements[residue.attrib[id]] = name
        if "type" not in residue.attrib:
            atoms = copy(all_residues)
        elif residue.attrib["type"] == "Protein":
            atoms = copy(protein_residues)
        elif residue.attrib["type"] == "Nucleic":
            atoms = copy(nucleic_acid_residues)
        else:
            atoms = copy(all_residues)
        _parse_residue(residue, atoms)
        atom_name_replacements[name] = atoms
    return residue_name_replacements, atom_name_replacements


def _parse_residue(residue, map):
    """Closely follows `mdtraj.formats.pdb.PDBTrajectoryFile._parseResidueAtoms`."""
    for atom in residue.findall("Atom"):
        name = atom.attrib["name"]
        for id in atom.attrib:
            map[atom.attrib[id]] = name


def _validate_pdb_file(filename):
    if filename.suffixes not in [[".pdb"], [".pdb", ".gz"]]:  # , [".cif"]]:
        raise ValueError(
            "PDB filename must have suffix `.pdb` or `.pdb.gz`"  # , or 'cif'. "
            f"Got filename {filename}."
        )


def _is_url(url):
    """Check to see if a URL has a valid protocol.
    Originally pandas/io.common.py, sourced from `mdtraj`.
    Copyright 2014 Pandas Developers, used under the BSD licence.
    """
    try:
        return urlparse(url).scheme in _VALID_URLS
    except (AttributeError, TypeError):
        return False


def _open_maybe_zipped(filename, mode, force_overwrite=True):
    """Open a file in text (not binary) mode, transparently handling
    .gz or .bz2 compresssion, with utf-8 encoding.
    Closely follows `mdtraj.utils.open_maybe_zipped`.
    """
    _, extension = os.path.splitext(str(filename).lower())
    if mode == "r":
        if extension == ".gz":
            with gzip.GzipFile(filename, "r") as gz_f:
                return StringIO(gz_f.read().decode("utf-8"))
        elif extension == ".bz2":
            with bz2.BZ2File(filename, "r") as bz2_f:
                return StringIO(bz2_f.read().decode("utf-8"))
        else:
            return open(filename)
    elif mode == "w":
        if os.path.exists(filename) and not force_overwrite:
            raise OSError('"%s" already exists' % filename)
        if extension == ".gz":
            binary_fh = gzip.GzipFile(filename, "wb")
            return io.TextIOWrapper(binary_fh, encoding="utf-8")
        elif extension == ".bz2":
            binary_fh = bz2.BZ2File(filename, "wb")
            return io.TextIOWrapper(binary_fh, encoding="utf-8")
        else:
            return open(filename, "w")
    else:
        raise Exception(f"Internal error opening file {filename}. Invalid mode {mode}.")



from ._base_distribution import (
    AbstractDistribution as AbstractDistribution,
)
from ._gaussian_distributions import (
    AbstractGaussianDistribution as AbstractGaussianDistribution,
    IndependentGaussianFourierModes as IndependentGaussianFourierModes,
    IndependentGaussianPixels as IndependentGaussianPixels,
)



from . import operators as operators, transforms as transforms
from ._average import (
    compute_binned_radial_average as compute_binned_radial_average,
    interpolate_radial_average_on_grid as interpolate_radial_average_on_grid,
)
from ._downsample import (
    downsample_to_shape_with_fourier_cropping as downsample_to_shape_with_fourier_cropping,  # noqa: E501
    downsample_with_fourier_cropping as downsample_with_fourier_cropping,
)
from ._edges import (
    crop_to_shape as crop_to_shape,
    crop_to_shape_with_center as crop_to_shape_with_center,
    pad_to_shape as pad_to_shape,
    resize_with_crop_or_pad as resize_with_crop_or_pad,
)
from ._fft import (
    fftn as fftn,
    ifftn as ifftn,
    irfftn as irfftn,
    rfftn as rfftn,
)
from ._fourier_statistics import (
    compute_binned_powerspectrum as compute_binned_powerspectrum,
    compute_fourier_ring_correlation as compute_fourier_ring_correlation,
    compute_fourier_shell_correlation as compute_fourier_shell_correlation,
)
from ._fourier_utils import (
    convert_fftn_to_rfftn as convert_fftn_to_rfftn,
    enforce_self_conjugate_rfftn_components as enforce_self_conjugate_rfftn_components,
)
from ._map_coordinates import (
    compute_spline_coefficients as compute_spline_coefficients,
    map_coordinates as map_coordinates,
    map_coordinates_with_cubic_spline as map_coordinates_with_cubic_spline,
)
from ._normalize import normalize_image as normalize_image, rescale_image as rescale_image
from ._rescale_pixel_size import (
    maybe_rescale_pixel_size as maybe_rescale_pixel_size,
    rescale_pixel_size as rescale_pixel_size,
)



import pytest

import cryojax.simulator as cxs
from cryojax.constants import (
    get_tabulated_scattering_factor_parameters,
    read_peng_element_scattering_factor_parameter_table,
)
from cryojax.io import read_array_from_mrc, read_atoms_from_pdb
from cryojax.simulator import DiscreteStructuralEnsemble


@pytest.fixture
def voxel_potential(sample_mrc_path):
    real_voxel_grid, voxel_size = read_array_from_mrc(sample_mrc_path, loads_spacing=True)
    return cxs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size, pad_scale=1.3
    )


@pytest.fixture
def atom_potential(sample_pdb_path):
    atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="not element H",
        loads_b_factors=True,
    )
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_identities, read_peng_element_scattering_factor_parameter_table()
    )
    return cxs.PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
        b_factors=b_factors,
    )


@pytest.mark.parametrize(
    "potential",
    [("voxel_potential"), ("atom_potential")],
)
def test_conformation(potential, request):
    potential = request.getfixturevalue(potential)
    pose = cxs.EulerAnglePose()
    conformational_space = tuple([potential for _ in range(3)])
    structure = DiscreteStructuralEnsemble(conformational_space, pose, conformation=0)
    _ = structure.get_potential_in_transformed_frame()



"""The image configuration and utility manager."""

import math
from functools import cached_property
from typing import Any, Callable, Optional, TypedDict

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float, Inexact

from ..constants import convert_keV_to_angstroms
from ..coordinates import make_coordinate_grid, make_frequency_grid
from ..internal import error_if_not_positive
from ..ndimage import (
    crop_to_shape,
    pad_to_shape,
    resize_with_crop_or_pad,
)


class GridHelper(eqx.Module, strict=True):
    coordinate_grid: Float[Array, "y_dim x_dim 2"]
    frequency_grid: Float[Array, "y_dim x_dim//2+1 2"]
    full_frequency_grid: Optional[Float[Array, "y_dim x_dim 2"]]

    def __init__(self, shape: tuple[int, int], only_rfft: bool = True):
        self.coordinate_grid = make_coordinate_grid(shape)
        self.frequency_grid = make_frequency_grid(shape, outputs_rfftfreqs=True)
        if only_rfft:
            self.full_frequency_grid = None
        else:
            self.full_frequency_grid = make_frequency_grid(shape, outputs_rfftfreqs=False)


class PadOptions(TypedDict):
    shape: tuple[int, int]
    grid_helper: Optional[GridHelper]
    mode: str | Callable


class AbstractConfig(eqx.Module, strict=True):
    """Configuration and utilities for an electron microscopy image."""

    shape: eqx.AbstractVar[tuple[int, int]]
    pixel_size: eqx.AbstractVar[Float[Array, ""]]
    voltage_in_kilovolts: eqx.AbstractVar[Float[Array, ""]]

    grid_helper: eqx.AbstractVar[Optional[GridHelper]]
    pad_options: eqx.AbstractVar[PadOptions]

    def __check_init__(self):
        if self.padded_shape[0] < self.shape[0] or self.padded_shape[1] < self.shape[1]:
            raise AttributeError(
                "ImageConfig.padded_shape is less than ImageConfig.shape in one or "
                "more dimensions."
            )

    @property
    def wavelength_in_angstroms(self) -> Float[Array, ""]:
        """The incident electron wavelength corresponding to the beam
        energy `voltage_in_kilovolts`.
        """
        return convert_keV_to_angstroms(self.voltage_in_kilovolts)

    @property
    def wavenumber_in_inverse_angstroms(self) -> Float[Array, ""]:
        """The incident electron wavenumber corresponding to the beam
        energy `voltage_in_kilovolts`.
        """
        return 2 * jnp.pi / self.wavelength_in_angstroms

    @cached_property
    def coordinate_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """A spatial coordinate system for the `shape`."""
        if self.grid_helper is None:
            return make_coordinate_grid(self.shape)
        else:
            return self.grid_helper.coordinate_grid

    @cached_property
    def coordinate_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """Convenience property for `pixel_size * coordinate_grid_in_pixels`"""
        return _safe_multiply_by_constant(self.coordinate_grid_in_pixels, self.pixel_size)

    @cached_property
    def frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim//2+1} 2"]:
        """A spatial frequency coordinate system for the `shape`,
        with hermitian symmetry.
        """
        if self.grid_helper is None:
            return make_frequency_grid(self.shape, outputs_rfftfreqs=True)
        else:
            return self.grid_helper.frequency_grid

    @cached_property
    def frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim//2+1} 2"]:
        """Convenience property for `frequency_grid_in_pixels / pixel_size`"""
        return _safe_multiply_by_constant(
            self.frequency_grid_in_pixels, 1 / self.pixel_size
        )

    @cached_property
    def full_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """A spatial frequency coordinate system for the `shape`,
        without hermitian symmetry.
        """
        if self.grid_helper is None or self.grid_helper.full_frequency_grid is None:
            return make_frequency_grid(shape=self.shape, outputs_rfftfreqs=False)
        else:
            return self.grid_helper.full_frequency_grid

    @cached_property
    def full_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """Convenience property for `full_frequency_grid_in_pixels / pixel_size`"""
        return _safe_multiply_by_constant(
            self.full_frequency_grid_in_pixels, 1 / self.pixel_size
        )

    @property
    def padded_shape(self):
        return self.pad_options["shape"]

    @property
    def pad_mode(self):
        return self.pad_options["mode"]

    @cached_property
    def padded_coordinate_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """A spatial coordinate system for the `padded_shape`."""
        grid_helper = (
            self.grid_helper
            if self.shape == self.padded_shape
            else self.pad_options["grid_helper"]
        )
        if grid_helper is None:
            return make_coordinate_grid(shape=self.padded_shape)
        else:
            return grid_helper.coordinate_grid

    @cached_property
    def padded_coordinate_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """Convenience property for `pixel_size * padded_coordinate_grid_in_pixels`"""
        return _safe_multiply_by_constant(
            self.padded_coordinate_grid_in_pixels, self.pixel_size
        )

    @cached_property
    def padded_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1} 2"]:
        """A spatial frequency coordinate system for the `padded_shape`,
        with hermitian symmetry.
        """
        grid_helper = (
            self.grid_helper
            if self.shape == self.padded_shape
            else self.pad_options["grid_helper"]
        )
        if grid_helper is None:
            return make_frequency_grid(shape=self.padded_shape, outputs_rfftfreqs=True)
        else:
            return grid_helper.frequency_grid

    @cached_property
    def padded_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1} 2"]:
        """Convenience property for `padded_frequency_grid_in_pixels / pixel_size`"""
        return _safe_multiply_by_constant(
            self.padded_frequency_grid_in_pixels, 1 / self.pixel_size
        )

    @cached_property
    def padded_full_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """A spatial frequency coordinate system for the `padded_shape`,
        without hermitian symmetry.
        """
        grid_helper = (
            self.grid_helper
            if self.shape == self.padded_shape
            else self.pad_options["grid_helper"]
        )
        if grid_helper is None or grid_helper.full_frequency_grid is None:
            return make_frequency_grid(shape=self.padded_shape, outputs_rfftfreqs=False)
        else:
            return grid_helper.full_frequency_grid

    @cached_property
    def padded_full_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """Convenience property for `padded_full_frequency_grid_in_pixels / pixel_size`"""
        return _safe_multiply_by_constant(
            self.padded_full_frequency_grid_in_pixels, 1 / self.pixel_size
        )

    def crop_to_shape(
        self, image: Inexact[Array, "y_dim x_dim"]
    ) -> Inexact[Array, "{self.y_dim} {self.x_dim}"]:
        """Crop an image to `shape`."""
        return crop_to_shape(image, self.shape)

    def pad_to_padded_shape(
        self, image: Inexact[Array, "y_dim x_dim"], **kwargs: Any
    ) -> Inexact[Array, "{self.padded_y_dim} {self.padded_x_dim}"]:
        """Pad an image to `padded_shape`."""
        return pad_to_shape(image, self.padded_shape, mode=self.pad_mode, **kwargs)

    def crop_or_pad_to_padded_shape(
        self, image: Inexact[Array, "y_dim x_dim"], **kwargs: Any
    ) -> Inexact[Array, "{self.padded_y_dim} {self.padded_x_dim}"]:
        """Reshape an image to `padded_shape` using cropping or padding."""
        return resize_with_crop_or_pad(
            image, self.padded_shape, mode=self.pad_mode, **kwargs
        )

    @property
    def n_pixels(self) -> int:
        """Convenience property for `math.prod(shape)`"""
        return math.prod(self.shape)

    @property
    def y_dim(self) -> int:
        """Convenience property for `shape[0]`"""
        return self.shape[0]

    @property
    def x_dim(self) -> int:
        """Convenience property for `shape[1]`"""
        return self.shape[1]

    @property
    def padded_y_dim(self) -> int:
        """Convenience property for `padded_shape[0]`"""
        return self.padded_shape[0]

    @property
    def padded_x_dim(self) -> int:
        """Convenience property for `padded_shape[1]`"""
        return self.padded_shape[1]

    @property
    def padded_n_pixels(self) -> int:
        """Convenience property for `math.prod(padded_shape)`"""
        return math.prod(self.padded_shape)


class BasicConfig(AbstractConfig, strict=True):
    """Configuration and utilities for a basic electron microscopy
    image."""

    shape: tuple[int, int] = eqx.field(static=True)
    pixel_size: Float[Array, ""]
    voltage_in_kilovolts: Float[Array, ""]

    grid_helper: Optional[GridHelper]
    pad_options: PadOptions

    def __init__(
        self,
        shape: tuple[int, int],
        pixel_size: float | Float[Array, ""],
        voltage_in_kilovolts: float | Float[Array, ""],
        *,
        grid_helper: Optional[GridHelper] = None,
        pad_options: dict[str, Any] = {},
    ):
        """**Arguments:**

        - `shape`:
            Shape of the imaging plane in pixels.
        - `pixel_size`:
            The pixel size of the image in angstroms.
        - `voltage_in_kilovolts`:
            The incident energy of the electron beam.
        - `electrons_per_angstrom_squared`:
            The integrated dose rate of the electron beam.
        - `grid_helper`:
            The `GridHelper` object, which stores the coordinate grids
            for image shape `shape`.
            If not passed, grid are computed at run-time.
        - `pad_options`:
            Options that control image padding.
            - `shape`:
                The shape of the image after padding. By default, equal
                to `shape`.
            - `mode`:
                The method of image padding. By default, equal
                to `"constant"`. For all options, see `jax.numpy.pad`.
            - `grid_helper`:
                The `GridHelper` object, which stores coordinate grids
                for the padded shape. If not passed, grid are computed
                at run-time. By default, equal to `None`.
        """
        # Set parameters
        self.pixel_size = error_if_not_positive(jnp.asarray(pixel_size, dtype=float))
        self.voltage_in_kilovolts = error_if_not_positive(
            jnp.asarray(voltage_in_kilovolts, dtype=float)
        )
        # Set shape
        self.shape = shape
        # Set pad options
        self.pad_options = _dict_to_pad_options(pad_options, shape)
        # Finally, grid helper
        self.grid_helper = grid_helper


class DoseConfig(AbstractConfig, strict=True):
    """Configuration and utilities for an electron microscopy image,
    including the electron dose."""

    shape: tuple[int, int] = eqx.field(static=True)
    pixel_size: Float[Array, ""]
    voltage_in_kilovolts: Float[Array, ""]
    electrons_per_angstrom_squared: Float[Array, ""]

    grid_helper: Optional[GridHelper]
    pad_options: PadOptions

    def __init__(
        self,
        shape: tuple[int, int],
        pixel_size: float | Float[Array, ""],
        voltage_in_kilovolts: float | Float[Array, ""],
        electrons_per_angstrom_squared: float | Float[Array, ""],
        *,
        grid_helper: Optional[GridHelper] = None,
        pad_options: dict[str, Any] = {},
    ):
        """**Arguments:**

        - `shape`:
            Shape of the imaging plane in pixels.
        - `pixel_size`:
            The pixel size of the image in angstroms.
        - `voltage_in_kilovolts`:
            The incident energy of the electron beam.
        - `electrons_per_angstrom_squared`:
            The integrated dose rate of the electron beam.
        - `grid_helper`:
            The `GridHelper` object, which stores the coordinate grids
            for image shape `shape`.
            If not passed, grid are computed at run-time.
        - `pad_options`:
            Options that control image padding.
            - `shape`:
                The shape of the image after padding. By default, equal
                to `shape`.
            - `mode`:
                The method of image padding. By default, equal
                to `"constant"`. For all options, see `jax.numpy.pad`.
            - `grid_helper`:
                The `GridHelper` object, which stores coordinate grids
                for the padded shape. If not passed, grid are computed
                at run-time. By default, equal to `None`.
        """
        # Set parameters
        self.pixel_size = error_if_not_positive(jnp.asarray(pixel_size, dtype=float))
        self.voltage_in_kilovolts = error_if_not_positive(
            jnp.asarray(voltage_in_kilovolts, dtype=float)
        )
        self.electrons_per_angstrom_squared = jnp.asarray(
            electrons_per_angstrom_squared, dtype=float
        )
        # Set shape
        self.shape = shape
        # Set pad options
        self.pad_options = _dict_to_pad_options(pad_options, shape)
        # Finally, grid helper
        self.grid_helper = grid_helper


def _safe_multiply_by_constant(
    grid: Float[Array, "y_dim x_dim 2"], constant: Float[Array, ""]
) -> Float[Array, "y_dim x_dim 2"]:
    """Multiplies a coordinate grid by a constant in a
    safe way for gradient computation.
    """
    return jnp.where(grid != 0.0, jnp.asarray(constant) * grid, 0.0)


def _dict_to_pad_options(d: dict[str, Any], default_shape: tuple[int, int]) -> PadOptions:
    shape = d["shape"] if "shape" in d else default_shape
    mode = d["mode"] if "mode" in d else "constant"
    grid_helper = d["grid_helper"] if "grid_helper" in d else None

    return PadOptions(shape=shape, mode=mode, grid_helper=grid_helper)



from .base_potential_integrator import (
    AbstractPotentialIntegrator as AbstractPotentialIntegrator,
)
from .direct_integrator import (
    AbstractDirectIntegrator as AbstractDirectIntegrator,
    AbstractDirectVoxelIntegrator as AbstractDirectVoxelIntegrator,
    EwaldSphereExtraction as EwaldSphereExtraction,
    FourierSliceExtraction as FourierSliceExtraction,
    GaussianMixtureProjection as GaussianMixtureProjection,
    NufftProjection as NufftProjection,
)
from .multislice_integrator import (
    AbstractMultisliceIntegrator as AbstractMultisliceIntegrator,
    FFTMultisliceIntegrator as FFTMultisliceIntegrator,
)



import equinox as eqx


class AbstractPotentialIntegrator(eqx.Module, strict=True):
    """Base class for a method of integrating a potential onto
    the exit plane.
    """

    requires_inverse_rotation: eqx.AbstractClassVar[bool]



from .base_scattering_theory import (
    AbstractScatteringTheory as AbstractScatteringTheory,
    AbstractWaveScatteringTheory as AbstractWaveScatteringTheory,
)
from .high_energy_scattering_theory import (
    HighEnergyScatteringTheory as HighEnergyScatteringTheory,
)
from .multislice_scattering_theory import (
    MultisliceScatteringTheory as MultisliceScatteringTheory,
)
from .weak_phase_scattering_theory import (
    AbstractWeakPhaseScatteringTheory as AbstractWeakPhaseScatteringTheory,
    WeakPhaseScatteringTheory as WeakPhaseScatteringTheory,
)



"""
These versions of the scipy function map_coordinates is modified from Louis Desdoigts's
version: https://github.com/LouisDesdoigts/jax/blob/cubic-spline-updated/jax/_src/scipy/ndimage.py.

This code was developed for the project [`dLux`](https://louisdesdoigts.github.io/dLux/).
"""

import functools
import itertools
import operator
from typing import List, Sequence, Tuple

import jax.numpy as jnp
import lineax as lx
from jax import lax, vmap
from jaxtyping import Array, ArrayLike


def map_coordinates(
    input: Array,
    coordinates: Sequence[Array],
    order: int,
    mode: str = "fill",
    cval: float | complex = 0.0,
):
    """
    Similar to `scipy.map_coordinates`, but diverges from the API.

    Adapted from Louis Desdoigts's [version of `jax.scipy.map_coordinates`](https://github.com/LouisDesdoigts/jax/blob/cubic-spline-updated/jax/_src/scipy/ndimage.py),
    which was developed for the project [`dLux`](https://louisdesdoigts.github.io/dLux/).

    Arguments
    ---------
    mode :
        Uses built-in JAX out-of-bounds indexing to determine how to
        extrapolate beyond boundaries.
        See https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.ndarray.at.html.
    """
    if order in [0, 1]:
        return _map_coordinates_nn_or_linear(input, coordinates, order, mode, cval)
    elif order == 3:
        coefficients = compute_spline_coefficients(input)
        return _map_coordinates_with_cubic_spline(coefficients, coordinates, mode, cval)
    else:
        raise NotImplementedError(f"map_coordinates does not support order={order}.")


def map_coordinates_with_cubic_spline(
    coefficients: Array,
    coordinates: Sequence[Array],
    mode: str = "fill",
    cval: float | complex = 0.0,
):
    """
    Similar to scipy.map_coordinates, but takes cubic spline coefficients as input.

    Adapted from https://github.com/LouisDesdoigts/jax/blob/cubic-spline-updated/jax/_src/scipy/ndimage.py,
    which was developed for the project [dLux](https://louisdesdoigts.github.io/dLux/).

    Arguments
    ---------
    mode :
        Uses built-in JAX out-of-bounds indexing to determine how to
        extrapolate beyond boundaries.
        See https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.ndarray.at.html.

    """
    return _map_coordinates_with_cubic_spline(coefficients, coordinates, mode, cval)


def compute_spline_coefficients(data: Array) -> Array:
    """Solve for the spline coefficients of an input array."""
    ndim = data.ndim
    for i in range(ndim):
        axis = ndim - i - 1
        A = _build_operator(data.shape[axis] - 2, dtype=data.dtype)
        fn = lambda x: _solve_coefficients(x, A)
        for j in range(ndim - 2, -1, -1):
            ax = int(j >= axis)
            fn = vmap(fn, ax, ax)
        data = fn(data)
    return data


def _map_coordinates_nn_or_linear(
    input: ArrayLike,
    coordinates: Sequence[ArrayLike],
    order: int,
    mode: str,
    cval: float | complex,
) -> Array:
    input_arr = jnp.asarray(input)
    coordinate_arrs = [jnp.asarray(c) for c in coordinates]

    if len(coordinates) != input_arr.ndim:
        raise ValueError(
            "coordinates must be a sequence of length input.ndim, but " "{} != {}".format(
                len(coordinates), input_arr.ndim
            )
        )

    if order == 0:
        interp_fun = _nearest_indices_and_weights
    elif order == 1:
        interp_fun = _linear_indices_and_weights
    else:
        raise NotImplementedError(
            "_map_coordinates_nn_or_linear requires order = 0 or 1."
        )

    interpolations_1d = []
    for coordinate in coordinate_arrs:
        interp_nodes = interp_fun(coordinate)
        interpolations_1d.append(interp_nodes)
    outputs = []
    for items in itertools.product(*interpolations_1d):
        indices, weights = _unzip2(items)
        contribution = input_arr.at[indices].get(mode=mode, fill_value=cval)
        interpolated = _nonempty_prod(weights) * contribution
        outputs.append(interpolated)
    result = _nonempty_sum(outputs)
    if jnp.issubdtype(input_arr.dtype, jnp.integer):
        result = _round_half_away_from_zero(result)
    return result.astype(input_arr.dtype)


def _map_coordinates_with_cubic_spline(
    coefficients: ArrayLike,
    coordinates: Sequence[ArrayLike],
    mode: str,
    cval: complex | float,
) -> Array:
    coefficients = jnp.asarray(coefficients)
    if len(coordinates) != coefficients.ndim:
        raise ValueError(
            "coordinates must be a sequence of length coefficients.ndim, but "
            "{} != {}".format(len(coordinates), coefficients.ndim)
        )
    # Stack coordinates along one axis
    coords = jnp.stack([jnp.asarray(c) for c in coordinates], axis=0)
    # vmap spline evaluate over coordinates and return
    fn = lambda coord: _spline_point(coefficients, coord, mode, cval)
    return vmap(fn)(coords.reshape(coefficients.ndim, -1).T).reshape(coords.shape[1:])


#
# Nearest neighbor and linear interpolation utilities
#
def _nonempty_prod(arrs: Sequence[Array]) -> Array:
    return functools.reduce(operator.mul, arrs)


def _nonempty_sum(arrs: Sequence[Array]) -> Array:
    return functools.reduce(operator.add, arrs)


def _round_half_away_from_zero(a: Array) -> Array:
    return a if jnp.issubdtype(a.dtype, jnp.integer) else lax.round(a)


def _nearest_indices_and_weights(
    coordinate: Array,
) -> List[Tuple[Array, ArrayLike]]:
    index = _round_half_away_from_zero(coordinate).astype(jnp.int32)
    weight = coordinate.dtype.type(1)
    return [(index, weight)]


def _linear_indices_and_weights(
    coordinate: Array,
) -> List[Tuple[Array, ArrayLike]]:
    lower = jnp.floor(coordinate)
    upper_weight = coordinate - lower
    lower_weight = 1 - upper_weight
    index = lower.astype(jnp.int32)
    return [(index, lower_weight), (index + 1, upper_weight)]


#
# Spline interpolation utilities
#
def _build_operator(
    n: int, dtype: jnp.dtype | None = None, diag_value: float = 4.0
) -> lx.TridiagonalLinearOperator:
    diagonal = jnp.full((n,), diag_value, dtype=dtype)
    lower_diagonal = jnp.full((n - 1,), 1.0, dtype=dtype)
    upper_diagonal = jnp.full((n - 1,), 1.0, dtype=dtype)
    return lx.TridiagonalLinearOperator(diagonal, lower_diagonal, upper_diagonal)


def _construct_vector(data: Array, c2: Array, cnp2: Array) -> Array:
    yvec = data[1:-1]
    first = data[1] - c2
    last = data[-2] - cnp2
    yvec = yvec.at[0].set(first)
    yvec = yvec.at[-1].set(last)
    return yvec


def _solve_coefficients(
    data: Array, operator: lx.TridiagonalLinearOperator, h=1
) -> Array:
    # Calcualte second and second last coefficients
    c2 = 1 / 6 * data[0]
    cnp2 = 1 / 6 * data[-1]

    # Solve for internal cofficients
    yvec = _construct_vector(data, c2, cnp2)
    solution = lx.linear_solve(operator, yvec)
    cs = solution.value

    # Calculate first and last coefficients
    c1 = 2 * c2 - cs[0]
    cnp3 = 2 * cnp2 - cs[-1]
    return jnp.concatenate([jnp.array([c1, c2]), cs, jnp.array([cnp2, cnp3])])


def _spline_basis(t: Array) -> Array:
    at = jnp.abs(t)
    fn1 = lambda t: (2 - t) ** 3
    fn2 = lambda t: 4 - 6 * t**2 + 3 * t**3
    return jnp.where(
        at >= 1,
        jnp.where(at <= 2, fn1(at), 0),  # type: ignore
        jnp.where(at <= 1, fn2(at), 0),  # type: ignore
    )


def _spline_value(
    coefficients: Array,
    coordinate: Array,
    index: Array,
    mode: str,
    cval: complex | float,
) -> Array:
    coefficient = coefficients.at[tuple(index)].get(mode=mode, fill_value=cval)
    fn = vmap(lambda x, i: _spline_basis(x - i + 1), (0, 0))
    return coefficient * fn(coordinate, index).prod()


def _spline_point(
    coefficients: Array,
    coordinate: Array,
    mode: str,
    cval: complex | float,
) -> Array:
    index_fn = lambda x: (jnp.arange(0, 4) + jnp.floor(x)).astype(int)
    index_vals = vmap(index_fn)(coordinate)
    indices = jnp.array(jnp.meshgrid(*index_vals, indexing="ij"))
    fn = lambda index: _spline_value(coefficients, coordinate, index, mode, cval)
    return vmap(fn)(indices.reshape(coefficients.ndim, -1).T).sum()


#
# Utilities
#
def _unzip2(xys):
    xs = []
    ys = []
    for x, y in xys:
        xs.append(x)
        ys.append(y)
    return tuple(xs), tuple(ys)



"""
Image formation models.
"""

from abc import abstractmethod
from typing import Optional, TypedDict
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Bool, Complex, Float, PRNGKeyArray

from ...ndimage import irfftn, rfftn
from ...ndimage.transforms import FilterLike, MaskLike
from .._config import AbstractConfig
from .._potential_integrator import AbstractDirectIntegrator
from .._structure import AbstractStructure
from .._transfer_theory import ContrastTransferTheory


RealImageArray = Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
FourierImageArray = Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
PaddedRealImageArray = Float[
    Array,
    "{self.config.padded_y_dim} " "{self.config.padded_x_dim}",
]
PaddedFourierImageArray = Complex[
    Array,
    "{self.config.padded_y_dim} " "{self.config.padded_x_dim//2+1}",
]

ImageArray = RealImageArray | FourierImageArray
PaddedImageArray = PaddedRealImageArray | PaddedFourierImageArray


class NormalizeOptions(TypedDict):
    applies_mask: bool
    use_mask: bool


class AbstractImageModel(eqx.Module, strict=True):
    """Base class for an image formation model.

    Call an `AbstractImageModel`'s `render` routine.
    """

    structure: eqx.AbstractVar[AbstractStructure]
    config: eqx.AbstractVar[AbstractConfig]

    normalizes_signal: eqx.AbstractVar[bool]
    signal_region: eqx.AbstractVar[Optional[Bool[Array, "_ _"]]]

    @abstractmethod
    def compute_fourier_image(
        self, rng_key: Optional[PRNGKeyArray] = None
    ) -> ImageArray | PaddedImageArray:
        """Render an image without postprocessing."""
        raise NotImplementedError

    def simulate(
        self,
        rng_key: Optional[PRNGKeyArray] = None,
        *,
        removes_padding: bool = True,
        outputs_real_space: bool = True,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> ImageArray | PaddedImageArray:
        """Render an image.

        **Arguments:**

        - `rng_key`:
            The random number generator key. If not passed, render an image
            with no stochasticity.
        - `removes_padding`:
            If `True`, return an image cropped to `BasicConfig.shape`.
            Otherwise, return an image at the `BasicConfig.padded_shape`.
            If `removes_padding = False`, the `AbstractImageModel.filter`
            and `AbstractImageModel.mask` are not applied, overriding
            the booleans `applies_mask` and `applies_filter`.
        - `outputs_real_space`:
            If `True`, return the image in real space.
        - `mask`:
            Optionally apply a mask to the image.
        - `filter`:
            Optionally apply a filter to the image.
        """
        fourier_image = self.compute_fourier_image(rng_key)

        return self._maybe_postprocess(
            fourier_image,
            removes_padding=removes_padding,
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        )

    def postprocess(
        self,
        fourier_image: PaddedFourierImageArray,
        *,
        outputs_real_space: bool = True,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> ImageArray:
        """Return an image postprocessed with filters, cropping, masking,
        and normalization in either real or fourier space.
        """
        config = self.config
        if (
            mask is None
            and config.padded_shape == config.shape
            and not self.normalizes_signal
        ):
            # ... if there are no masks, we don't need to crop, and we are
            # not normalizing, minimize moving back and forth between real
            # and fourier space
            if filter is not None:
                fourier_image = filter(fourier_image)
            return (
                irfftn(fourier_image, s=config.shape)
                if outputs_real_space
                else fourier_image
            )
        else:
            # ... otherwise, apply filter, crop, and mask, again trying to
            # minimize moving back and forth between real and fourier space
            padded_rfft_shape = config.padded_frequency_grid_in_pixels.shape[0:2]
            if filter is not None:
                # ... apply the filter
                if not filter.array.shape == padded_rfft_shape:
                    raise ValueError(
                        "Found that the `filter` was shape "
                        f"{filter.array.shape}, but expected it to be "
                        f"shape {padded_rfft_shape}. You may have passed a "
                        f"fitler according to the `AbstractImageModel.config.shape`, "
                        "when the `AbstractImageModel.config.padded_shape` was expected."
                    )
                fourier_image = filter(fourier_image)
            image = irfftn(fourier_image, s=config.padded_shape)
            if config.padded_shape != config.shape:
                image = config.crop_to_shape(image)
            if self.normalizes_signal:
                image = self._normalize_image(image)
            if mask is not None:
                image = mask(image)
            return image if outputs_real_space else rfftn(image)

    def _apply_translation(
        self, fourier_image: PaddedFourierImageArray
    ) -> PaddedFourierImageArray:
        pose = self.structure.pose
        phase_shifts = pose.compute_translation_operator(
            self.config.padded_frequency_grid_in_angstroms
        )
        fourier_image = pose.translate_image(
            fourier_image,
            phase_shifts,
            self.config.padded_shape,
        )

        return fourier_image

    def _normalize_image(self, image: RealImageArray) -> RealImageArray:
        mean, std = (
            jnp.mean(image, where=self.signal_region),
            jnp.std(image, where=self.signal_region),
        )
        image = (image - mean) / std

        return image

    def _maybe_postprocess(
        self,
        image: PaddedFourierImageArray,
        *,
        removes_padding: bool = True,
        outputs_real_space: bool = True,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> PaddedImageArray | ImageArray:
        config = self.config
        if removes_padding:
            return self.postprocess(
                image, outputs_real_space=outputs_real_space, mask=mask, filter=filter
            )
        else:
            return irfftn(image, s=config.padded_shape) if outputs_real_space else image


class LinearImageModel(AbstractImageModel, strict=True):
    """An simple image model in linear image formation theory."""

    structure: AbstractStructure
    integrator: AbstractDirectIntegrator
    transfer_theory: ContrastTransferTheory
    config: AbstractConfig

    normalizes_signal: bool
    signal_region: Optional[Bool[Array, "_ _"]]

    def __init__(
        self,
        structure: AbstractStructure,
        config: AbstractConfig,
        integrator: AbstractDirectIntegrator,
        transfer_theory: ContrastTransferTheory,
        *,
        normalizes_signal: bool = False,
        signal_region: Optional[Bool[Array, "_ _"]] = None,
    ):
        """**Arguments:**

        - `structure`:
            The biological structure.
        - `config`:
            The configuration of the instrument, such as for the pixel size
            and the wavelength.
        - `integrator`: The method for integrating the scattering potential.
        - `transfer_theory`: The contrast transfer theory.
        - `normalizes_signal`:
            If `True`, normalizes_signal the image before returning.
        - `signal_region`:
            A boolean array that is 1 where there is signal,
            and 0 otherwise used to normalize the image.
            Must have shape equal to `AbstractImageModel.shape`.
        """
        # Simulator components
        self.config = config
        self.integrator = integrator
        self.structure = structure
        self.transfer_theory = transfer_theory
        # Options
        self.normalizes_signal = normalizes_signal
        self.signal_region = signal_region

    @override
    def compute_fourier_image(
        self, rng_key: Optional[PRNGKeyArray] = None
    ) -> ImageArray | PaddedImageArray:
        # Get potential in the lab frame
        potential = self.structure.get_potential_in_transformed_frame(
            apply_translation=False,
            apply_inverse_rotation=self.integrator.requires_inverse_rotation,
        )
        # Compute the projection image
        fourier_image = self.integrator.integrate(
            potential, self.config, outputs_real_space=False
        )
        # Compute the image
        fourier_image = self.transfer_theory.propagate_object(  # noqa: E501
            fourier_image,
            self.config,
            is_projection_approximation=self.integrator.is_projection_approximation,
            defocus_offset=self.structure.pose.offset_z_in_angstroms,
        )
        # Now for the in-plane translation
        fourier_image = self._apply_translation(fourier_image)

        return fourier_image


class ProjectionImageModel(AbstractImageModel, strict=True):
    """An simple image model for computing a projection."""

    structure: AbstractStructure
    integrator: AbstractDirectIntegrator
    config: AbstractConfig

    normalizes_signal: bool
    signal_region: Optional[Bool[Array, "_ _"]]

    def __init__(
        self,
        structure: AbstractStructure,
        config: AbstractConfig,
        integrator: AbstractDirectIntegrator,
        *,
        normalizes_signal: bool = False,
        signal_region: Optional[Bool[Array, "_ _"]] = None,
    ):
        """**Arguments:**

        - `structure`:
            The biological structure.
        - `config`:
            The configuration of the instrument, such as for the pixel size
            and the wavelength.
        - `integrator`: The method for integrating the scattering potential.
        - `normalizes_signal`:
            If `True`, normalizes_signal the image before returning.
        - `signal_region`:
            A boolean array that is 1 where there is signal,
            and 0 otherwise used to normalize the image.
            Must have shape equal to `AbstractImageModel.shape`.
        """
        # Simulator components
        self.config = config
        self.integrator = integrator
        self.structure = structure
        # Options
        self.normalizes_signal = normalizes_signal
        self.signal_region = signal_region

    @override
    def compute_fourier_image(
        self, rng_key: Optional[PRNGKeyArray] = None
    ) -> ImageArray | PaddedImageArray:
        # Get potential in the lab frame
        potential = self.structure.get_potential_in_transformed_frame(
            apply_translation=False,
            apply_inverse_rotation=self.integrator.requires_inverse_rotation,
        )
        # Compute the projection image
        fourier_image = self.integrator.integrate(
            potential, self.config, outputs_real_space=False
        )
        # Now for the in-plane translation
        fourier_image = self._apply_translation(fourier_image)

        return fourier_image



"""
Using the fourier slice theorem for computing volume projections.
"""

from typing import ClassVar, Optional
from typing_extensions import override

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ....ndimage import (
    convert_fftn_to_rfftn,
    fftn,
    ifftn,
    irfftn,
    map_coordinates,
    map_coordinates_with_cubic_spline,
    rfftn,
)
from ....ndimage.transforms import InverseSincMask
from ..._config import AbstractConfig
from ..._potential_representation import (
    FourierVoxelGridPotential,
    FourierVoxelSplinePotential,
)
from .base_direct_integrator import AbstractDirectVoxelIntegrator


class FourierSliceExtraction(AbstractDirectVoxelIntegrator, strict=True):
    """Integrate points to the exit plane using the Fourier projection-slice theorem.

    This extracts slices using interpolation methods housed in
    `cryojax.image.map_coordinates` and `cryojax.image.map_coordinates_with_cubic_spline`.
    """

    pixel_rescaling_mode: Optional[str]
    correction_mask: Optional[InverseSincMask]
    out_of_bounds_mode: str
    fill_value: complex

    is_projection_approximation: ClassVar[bool] = True
    requires_inverse_rotation: ClassVar[bool] = True

    def __init__(
        self,
        *,
        pixel_rescaling_mode: Optional[str] = None,
        correction_mask: Optional[InverseSincMask] = None,
        out_of_bounds_mode: str = "fill",
        fill_value: complex = 0.0 + 0.0j,
    ):
        """**Arguments:**

        - `pixel_rescaling_mode`:
            Method for rescaling the final image to the `AbstractConfig`
            pixel size. See `cryojax.image.rescale_pixel_size` for documentation.
        - `correction_mask`:
            A `cryojax.image.operators.SincCorrectionMask` for performing
            sinc-correction on the linear-interpolated projections. This
            should be computed on a coordinate grid with shape matching
            the `FourierVoxelGridPotential.shape`.
        - `out_of_bounds_mode`:
            Specify how to handle out of bounds indexing. See
            `cryojax.image.map_coordinates` for documentation.
        - `fill_value`:
            Value for filling out-of-bounds indices. Used only when
            `out_of_bounds_mode = "fill"`.
        """
        self.pixel_rescaling_mode = pixel_rescaling_mode
        self.correction_mask = correction_mask
        self.out_of_bounds_mode = out_of_bounds_mode
        self.fill_value = fill_value

    @override
    def integrate(
        self,
        potential: FourierVoxelGridPotential | FourierVoxelSplinePotential,
        config: AbstractConfig,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim//2+1}",
        ]
        | Float[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
    ):
        """Compute the integrated scattering potential at the `AbstractConfig` settings
        of a voxel-based representation in fourier-space, using fourier slice extraction.

        **Arguments:**

        - `potential`: The scattering potential representation.
        - `config`: The configuration of the resulting image.

        **Returns:**

        The extracted fourier voxels of the `potential`, at the
        `config.padded_shape` and the `config.pixel_size`.
        """
        frequency_slice = potential.frequency_slice_in_pixels
        N = frequency_slice.shape[1]
        if potential.shape != (N, N, N):
            raise AttributeError(
                "Only cubic boxes are supported for fourier slice extraction."
            )
        # Compute the fourier projection
        if isinstance(potential, FourierVoxelSplinePotential):
            fourier_in_plane_potential = (
                self.extract_fourier_slice_from_spline_coefficients(
                    potential.spline_coefficients,
                    frequency_slice,
                )
            )
        elif isinstance(potential, FourierVoxelGridPotential):
            fourier_in_plane_potential = self.extract_fourier_slice_from_grid_points(
                potential.fourier_voxel_grid,
                frequency_slice,
            )
        else:
            raise ValueError(
                "Supported types for `potential` are `FourierVoxelGridPotential` and "
                "`FourierVoxelSplinePotential`."
            )

        # Resize the image to match the AbstractConfig.padded_shape
        if config.padded_shape != (N, N):
            fourier_in_plane_potential = rfftn(
                config.crop_or_pad_to_padded_shape(
                    irfftn(fourier_in_plane_potential, s=(N, N))
                )
            )
        # Scale by voxel size to convert from projection to integral
        fourier_in_plane_potential *= potential.voxel_size
        # Re-scale to correct pixel size if its different
        fourier_in_plane_potential = self._maybe_rescale_pixel_size(
            fourier_in_plane_potential, potential, config, input_is_rfft=True
        )
        return (
            irfftn(fourier_in_plane_potential, s=config.padded_shape)
            if outputs_real_space
            else fourier_in_plane_potential
        )

    def extract_fourier_slice_from_spline_coefficients(
        self,
        spline_coefficients: Complex[Array, "coeff_dim coeff_dim coeff_dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
    ) -> Complex[Array, "dim dim//2+1"]:
        """Extract a fourier slice using the interpolation defined by
        `spline_coefficients` at coordinates `frequency_slice_in_pixels`.

        **Arguments:**

        - `spline_coefficients`:
            Spline coefficients of the density grid in fourier space.
            The coefficients should be computed from a `fourier_voxel_grid`
            with the zero frequency component in the center. These are
            typically computed with the function
            `cryojax.image.compute_spline_coefficients`.
        - `frequency_slice_in_pixels`:
            Frequency central slice coordinate system. The zero
            frequency component should be in the center.
        - `voxel_size`:
            The voxel size of the `fourier_voxel_grid`. This argument is
            not used in the `FourierSliceExtraction` class.
        - `wavelength_in_angstroms`:
            The wavelength of the incident electron beam. This argument is
            not used in the `FourierSliceExtraction` class.

        **Returns:**

        The interpolated fourier slice at coordinates `frequency_slice_in_pixels`.
        """
        return _extract_slice_with_cubic_spline(
            spline_coefficients,
            frequency_slice_in_pixels,
            mode=self.out_of_bounds_mode,
            cval=self.fill_value,
        )

    def extract_fourier_slice_from_grid_points(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
    ) -> Complex[Array, "dim dim//2+1"]:
        """Extract a fourier slice of the `fourier_voxel_grid` at coordinates
        `frequency_slice_in_pixels`.

        **Arguments:**

        - `fourier_voxel_grid`:
            Density grid in fourier space. The zero frequency component
            should be in the center.
        - `frequency_slice_in_pixels`:
            Frequency central slice coordinate system. The zero
            frequency component should be in the center.
        - `voxel_size`:
            The voxel size of the `fourier_voxel_grid`. This argument is
            not used in the `FourierSliceExtraction` class.
        - `wavelength_in_angstroms`:
            The wavelength of the incident electron beam. This argument is
            not used in the `FourierSliceExtraction` class.

        **Returns:**

        The interpolated fourier slice at coordinates `frequency_slice_in_pixels`.
        """
        fourier_slice = _extract_slice(
            fourier_voxel_grid,
            frequency_slice_in_pixels,
            interpolation_order=1,
            mode=self.out_of_bounds_mode,
            cval=self.fill_value,
        )
        if self.correction_mask is not None:
            fourier_slice = fftn(self.correction_mask(ifftn(fourier_slice)))

        return fourier_slice


class EwaldSphereExtraction(AbstractDirectVoxelIntegrator, strict=True):
    """Integrate points to the exit plane by extracting a surface of the ewald
    sphere in fourier space.

    This extracts surfaces using interpolation methods housed in
    `cryojax.image.map_coordinates` and `cryojax.image.map_coordinates_with_cubic_spline`.
    """

    pixel_rescaling_mode: Optional[str]
    correction_mask: Optional[InverseSincMask]
    out_of_bounds_mode: str
    fill_value: complex

    is_projection_approximation: ClassVar[bool] = False
    requires_inverse_rotation: ClassVar[bool] = True

    def __init__(
        self,
        *,
        pixel_rescaling_mode: Optional[str] = None,
        correction_mask: Optional[InverseSincMask] = None,
        out_of_bounds_mode: str = "fill",
        fill_value: complex = 0.0 + 0.0j,
    ):
        """**Arguments:**

        - `pixel_rescaling_mode`:
            Method for rescaling the final image to the `AbstractConfig`
            pixel size. See `cryojax.image.rescale_pixel_size` for documentation.
        - `correction_mask`:
            A `cryojax.image.operators.SincCorrectionMask` for performing
            sinc-correction on the linear-interpolated projections. This
            should be computed on a coordinate grid with shape matching
            the `FourierVoxelGridPotential.shape`.
        - `out_of_bounds_mode`:
            Specify how to handle out of bounds indexing. See
            `cryojax.image.map_coordinates` for documentation.
        - `fill_value`:
            Value for filling out-of-bounds indices. Used only when
            `out_of_bounds_mode = "fill"`.
        """
        self.pixel_rescaling_mode = pixel_rescaling_mode
        self.correction_mask = correction_mask
        self.out_of_bounds_mode = out_of_bounds_mode
        self.fill_value = fill_value

    @override
    def integrate(
        self,
        potential: FourierVoxelGridPotential | FourierVoxelSplinePotential,
        config: AbstractConfig,
        outputs_real_space: bool = False,
    ) -> (
        Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
        | Float[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
    ):
        """Compute the integrated scattering potential at the `AbstractConfig` settings
        of a voxel-based representation in fourier-space, using fourier slice extraction.

        **Arguments:**

        - `potential`: The scattering potential representation.
        - `config`: The configuration of the resulting image.

        **Returns:**

        The extracted fourier voxels of the `potential`, at the
        `config.padded_shape` and the `config.pixel_size`.
        """
        frequency_slice = potential.frequency_slice_in_pixels
        N = frequency_slice.shape[1]
        if potential.shape != (N, N, N):
            raise AttributeError(
                "Only cubic boxes are supported for fourier slice extraction."
            )
        # Compute the fourier projection
        if isinstance(potential, FourierVoxelSplinePotential):
            ewald_sphere_surface = self.extract_ewald_sphere_from_spline_coefficients(
                potential.spline_coefficients,
                frequency_slice,
                potential.voxel_size,
                config.wavelength_in_angstroms,
            )
        elif isinstance(potential, FourierVoxelGridPotential):
            ewald_sphere_surface = self.extract_ewald_sphere_from_grid_points(
                potential.fourier_voxel_grid,
                frequency_slice,
                potential.voxel_size,
                config.wavelength_in_angstroms,
            )
        else:
            raise ValueError(
                "Supported types for `potential` are `FourierVoxelGridPotential` and "
                "`FourierVoxelSplinePotential`."
            )

        # Resize the image to match the AbstractConfig.padded_shape
        if config.padded_shape != (N, N):
            ewald_sphere_surface = fftn(
                config.crop_or_pad_to_padded_shape(ifftn(ewald_sphere_surface, s=(N, N)))
            )
        # Scale by voxel size to convert from projection to integral
        ewald_sphere_surface *= potential.voxel_size
        # Re-scale to correct pixel size if its different
        ewald_sphere_surface = self._maybe_rescale_pixel_size(
            ewald_sphere_surface, potential, config, input_is_rfft=False
        )
        return (
            irfftn(ewald_sphere_surface, s=config.padded_shape)
            if outputs_real_space
            else ewald_sphere_surface
        )

    def extract_ewald_sphere_from_spline_coefficients(
        self,
        spline_coefficients: Complex[Array, "coeff_dim coeff_dim coeff_dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""],
        wavelength_in_angstroms: Float[Array, ""],
    ) -> Complex[Array, "dim dim"]:
        """Extract an ewald sphere surface of the `fourier_voxel_grid` at
        coordinates normal to `frequency_slice_in_pixels` at wavelength
        `wavelength_in_angstroms`.

        **Arguments:**

        - `spline_coefficients`:
            Spline coefficients of the density grid in fourier space.
            The coefficients should be computed from a `fourier_voxel_grid`
            with the zero frequency component in the center. These are
            typically computed with the function
            `cryojax.image.compute_spline_coefficients`.
        - `frequency_slice_in_pixels`:
            Frequency central slice coordinate system. The zero
            frequency component should be in the center.
        - `voxel_size`:
            The voxel size of the `fourier_voxel_grid`.
        - `wavelength_in_angstroms`:
            The wavelength of the incident electron beam.

        **Returns:**

        The interpolated ewald sphere surface at coordinates normal to
        `frequency_slice_in_pixels`.
        """
        return _extract_ewald_sphere_surface_with_cubic_spline(
            spline_coefficients,
            frequency_slice_in_pixels,
            voxel_size,
            wavelength_in_angstroms,
            mode=self.out_of_bounds_mode,
            cval=self.fill_value,
        )

    def extract_ewald_sphere_from_grid_points(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""],
        wavelength_in_angstroms: Float[Array, ""],
    ) -> Complex[Array, "dim dim"]:
        """Extract an ewald sphere surface of the `fourier_voxel_grid` at
        coordinates normal to `frequency_slice_in_pixels` at wavelength
        `wavelength_in_angstroms`.

        **Arguments:**

        - `fourier_voxel_grid`:
            Density grid in fourier space. The zero frequency component
            should be in the center.
        - `frequency_slice_in_pixels`:
            Frequency central slice coordinate system. The zero
            frequency component should be in the center.
        - `voxel_size`:
            The voxel size of the `fourier_voxel_grid`.
        - `wavelength_in_angstroms`:
            The wavelength of the incident electron beam.

        **Returns:**

        The interpolated ewald sphere surface at coordinates normal to
        `frequency_slice_in_pixels`.
        """
        ewald_sphere_surface = _extract_ewald_sphere_surface(
            fourier_voxel_grid,
            frequency_slice_in_pixels,
            voxel_size,
            wavelength_in_angstroms,
            interpolation_order=1,
            mode=self.out_of_bounds_mode,
            cval=self.fill_value,
        )
        if self.correction_mask is not None:
            ewald_sphere_surface = fftn(self.correction_mask(ifftn(ewald_sphere_surface)))

        return ewald_sphere_surface


def _extract_slice(
    fourier_voxel_grid,
    frequency_slice,
    interpolation_order,
    **kwargs,
) -> Complex[Array, "dim dim//2+1"]:
    return convert_fftn_to_rfftn(
        _extract_surface_from_voxel_grid(
            fourier_voxel_grid,
            frequency_slice,
            is_spline_coefficients=False,
            interpolation_order=interpolation_order,
            **kwargs,
        ),
        mode="real",
    )


def _extract_slice_with_cubic_spline(
    spline_coefficients, frequency_slice, **kwargs
) -> Complex[Array, "dim dim//2+1"]:
    return convert_fftn_to_rfftn(
        _extract_surface_from_voxel_grid(
            spline_coefficients, frequency_slice, is_spline_coefficients=True, **kwargs
        ),
        mode="real",
    )


def _extract_ewald_sphere_surface(
    fourier_voxel_grid,
    frequency_slice,
    voxel_size,
    wavelength,
    interpolation_order,
    **kwargs,
) -> Complex[Array, "dim dim"]:
    ewald_sphere_frequencies = _get_ewald_sphere_surface_from_slice(
        frequency_slice, voxel_size, wavelength
    )
    return _extract_surface_from_voxel_grid(
        fourier_voxel_grid,
        ewald_sphere_frequencies,
        is_spline_coefficients=False,
        interpolation_order=interpolation_order,
        **kwargs,
    )


def _extract_ewald_sphere_surface_with_cubic_spline(
    spline_coefficients, frequency_slice, voxel_size, wavelength, **kwargs
) -> Complex[Array, "dim dim"]:
    ewald_sphere_frequencies = _get_ewald_sphere_surface_from_slice(
        frequency_slice, voxel_size, wavelength
    )
    return _extract_surface_from_voxel_grid(
        spline_coefficients,
        ewald_sphere_frequencies,
        is_spline_coefficients=True,
        **kwargs,
    )


def _get_ewald_sphere_surface_from_slice(
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
    voxel_size: Float[Array, ""],
    wavelength: Float[Array, ""],
) -> Float[Array, "1 dim dim 3"]:
    frequency_slice_with_zero_in_corner = jnp.fft.ifftshift(
        frequency_slice_in_pixels, axes=(0, 1, 2)
    )
    # Get zhat unit vector of the frequency slice
    xhat, yhat = (
        frequency_slice_with_zero_in_corner[0, 0, 1, :],
        frequency_slice_with_zero_in_corner[0, 1, 0, :],
    )
    xhat, yhat = xhat / jnp.linalg.norm(xhat), yhat / jnp.linalg.norm(yhat)
    zhat = jnp.cross(xhat, yhat)
    # Compute the ewald sphere surface, assuming the frequency slice is
    # in a rotated frame
    q_at_slice = frequency_slice_in_pixels
    q_squared = jnp.sum(q_at_slice**2, axis=-1)
    q_at_surface = (
        q_at_slice
        + (wavelength / voxel_size)
        * (q_squared[..., None] * zhat[None, None, None, :])
        / 2
    )
    return q_at_surface


def _extract_surface_from_voxel_grid(
    voxel_grid,
    frequency_coordinates,
    is_spline_coefficients=False,
    interpolation_order=1,
    **kwargs,
):
    # Convert to logical coordinates
    N = frequency_coordinates.shape[1]
    logical_frequency_coordinates = (frequency_coordinates * N) + N // 2
    # Convert arguments to map_coordinates convention and compute
    k_z, k_y, k_x = jnp.transpose(logical_frequency_coordinates, axes=[3, 0, 1, 2])
    if is_spline_coefficients:
        spline_coefficients = voxel_grid
        surface = map_coordinates_with_cubic_spline(
            spline_coefficients, (k_x, k_y, k_z), **kwargs
        )[0, :, :]
    else:
        fourier_voxel_grid = voxel_grid
        surface = map_coordinates(
            fourier_voxel_grid, (k_x, k_y, k_z), interpolation_order, **kwargs
        )[0, :, :]
    # Shift zero frequency component to corner
    surface = jnp.fft.ifftshift(surface)

    return surface



from .atom_potential import (
    AbstractAtomicPotential as AbstractAtomicPotential,
    AbstractTabulatedAtomicPotential as AbstractTabulatedAtomicPotential,
    GaussianMixtureAtomicPotential as GaussianMixtureAtomicPotential,
    PengAtomicPotential as PengAtomicPotential,
)
from .base_potential import (
    AbstractPotentialRepresentation as AbstractPotentialRepresentation,
)
from .voxel_potential import (
    AbstractFourierVoxelGridPotential as AbstractFourierVoxelGridPotential,
    AbstractVoxelPotential as AbstractVoxelPotential,
    FourierVoxelGridPotential as FourierVoxelGridPotential,
    FourierVoxelSplinePotential as FourierVoxelSplinePotential,
    RealVoxelCloudPotential as RealVoxelCloudPotential,
    RealVoxelGridPotential as RealVoxelGridPotential,
)
